"""Functions related to parameter sampling"""

from typing import Sequence, Tuple

import numpy as np
import pandas as pd

from . import parameters
from .C import *  # noqa: F403

__all__ = ["sample_from_prior", "sample_parameter_startpoints"]


def sample_from_prior(
    prior: Tuple[str, list, str, list], n_starts: int
) -> np.array:
    """Creates samples for one parameter based on prior

    Arguments:
        prior: A tuple as obtained from
            :func:`petab.parameter.get_priors_from_df`
        n_starts: Number of samples

    Returns:
        Array with sampled values
    """
    # unpack info
    p_type, p_params, scaling, bounds = prior

    # define a function to rescale the sampled points to parameter scale
    def scale(x):
        if scaling == LIN:
            return x
        if scaling == LOG:
            return np.log(x)
        if scaling == LOG10:
            return np.log10(x)
        raise NotImplementedError(
            f"Parameter priors on the parameter scale {scaling} are "
            "currently not implemented."
        )

    def clip_to_bounds(x: np.array):
        """Clip values in array x to bounds"""
        return np.maximum(np.minimum(scale(bounds[1]), x), scale(bounds[0]))

    # define lambda functions for each parameter
    if p_type == UNIFORM:
        sp = scale(
            (p_params[1] - p_params[0]) * np.random.random((n_starts,))
            + p_params[0]
        )

    elif p_type == PARAMETER_SCALE_UNIFORM:
        sp = (p_params[1] - p_params[0]) * np.random.random(
            (n_starts,)
        ) + p_params[0]

    elif p_type == NORMAL:
        sp = scale(
            np.random.normal(
                loc=p_params[0], scale=p_params[1], size=(n_starts,)
            )
        )

    elif p_type == LOG_NORMAL:
        sp = scale(
            np.exp(
                np.random.normal(
                    loc=p_params[0], scale=p_params[1], size=(n_starts,)
                )
            )
        )

    elif p_type == PARAMETER_SCALE_NORMAL:
        sp = np.random.normal(
            loc=p_params[0], scale=p_params[1], size=(n_starts,)
        )

    elif p_type == LAPLACE:
        sp = scale(
            np.random.laplace(
                loc=p_params[0], scale=p_params[1], size=(n_starts,)
            )
        )

    elif p_type == LOG_LAPLACE:
        sp = scale(
            np.exp(
                np.random.laplace(
                    loc=p_params[0], scale=p_params[1], size=(n_starts,)
                )
            )
        )

    elif p_type == PARAMETER_SCALE_LAPLACE:
        sp = np.random.laplace(
            loc=p_params[0], scale=p_params[1], size=(n_starts,)
        )

    else:
        raise NotImplementedError(
            f"Parameter priors of type {prior[0]} are not implemented."
        )

    return clip_to_bounds(sp)


def sample_parameter_startpoints(
    parameter_df: pd.DataFrame,
    n_starts: int = 100,
    seed: int = None,
    parameter_ids: Sequence[str] = None,
) -> np.array:
    """Create :class:`numpy.array` with starting points for an optimization

    Arguments:
        parameter_df: PEtab parameter DataFrame
        n_starts: Number of points to be sampled
        seed: Random number generator seed (see :func:`numpy.random.seed`)
        parameter_ids: A sequence of parameter IDs for which to sample starting points.
            For subsetting or reordering the parameters.
            Defaults to all estimated parameters.

    Returns:
        Array of sampled starting points with dimensions
        `n_startpoints` x `n_optimization_parameters`
    """
    if seed is not None:
        np.random.seed(seed)

    # get types and parameters of priors from dataframe
    prior_list = parameters.get_priors_from_df(
        parameter_df, mode=INITIALIZATION, parameter_ids=parameter_ids
    )

    startpoints = [sample_from_prior(prior, n_starts) for prior in prior_list]

    return np.array(startpoints).T



"""Functions operating on the PEtab parameter table"""

import numbers
import warnings
from collections import OrderedDict
from pathlib import Path
from typing import (
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import libsbml
import numpy as np
import pandas as pd

from . import conditions, core, lint, measurements, observables
from .C import *  # noqa: F403
from .models import Model

__all__ = [
    "create_parameter_df",
    "get_optimization_parameter_scaling",
    "get_optimization_parameters",
    "get_parameter_df",
    "get_priors_from_df",
    "get_valid_parameters_for_parameter_table",
    "map_scale",
    "map_unscale",
    "normalize_parameter_df",
    "scale",
    "unscale",
    "write_parameter_df",
]

PARAMETER_SCALE_ARGS = Literal["", "lin", "log", "log10"]


def get_parameter_df(
    parameter_file: Union[
        str, Path, pd.DataFrame, Iterable[Union[str, Path, pd.DataFrame]], None
    ]
) -> Union[pd.DataFrame, None]:
    """
    Read the provided parameter file into a ``pandas.Dataframe``.

    Arguments:
        parameter_file: Name of the file to read from or pandas.Dataframe,
        or an Iterable.

    Returns:
        Parameter ``DataFrame``, or ``None`` if ``None`` was passed.
    """
    if parameter_file is None:
        return None
    if isinstance(parameter_file, pd.DataFrame):
        parameter_df = parameter_file
    elif isinstance(parameter_file, (str, Path)):
        parameter_df = pd.read_csv(
            parameter_file, sep="\t", float_precision="round_trip"
        )
    elif isinstance(parameter_file, Iterable):
        dfs = [get_parameter_df(x) for x in parameter_file if x]

        if not dfs:
            return None

        parameter_df = pd.concat(dfs)
        # Check for contradicting parameter definitions
        _check_for_contradicting_parameter_definitions(parameter_df)

        return parameter_df

    lint.assert_no_leading_trailing_whitespace(
        parameter_df.columns.values, "parameter"
    )

    if not isinstance(parameter_df.index, pd.RangeIndex):
        parameter_df.reset_index(inplace=True)

    try:
        parameter_df.set_index([PARAMETER_ID], inplace=True)
    except KeyError as e:
        raise KeyError(
            f"Parameter table missing mandatory field {PARAMETER_ID}."
        ) from e
    _check_for_contradicting_parameter_definitions(parameter_df)

    return parameter_df


def _check_for_contradicting_parameter_definitions(parameter_df: pd.DataFrame):
    """
    Raises a ValueError for non-unique parameter IDs
    """
    parameter_duplicates = set(
        parameter_df.index.values[parameter_df.index.duplicated()]
    )
    if parameter_duplicates:
        raise ValueError(
            f"The values of `{PARAMETER_ID}` must be unique. The "
            f"following duplicates were found:\n{parameter_duplicates}"
        )


def write_parameter_df(df: pd.DataFrame, filename: Union[str, Path]) -> None:
    """Write PEtab parameter table

    Arguments:
        df: PEtab parameter table
        filename: Destination file name
    """
    df = get_parameter_df(df)
    df.to_csv(filename, sep="\t", index=True)


def get_optimization_parameters(parameter_df: pd.DataFrame) -> List[str]:
    """
    Get list of optimization parameter IDs from parameter table.

    Arguments:
        parameter_df: PEtab parameter DataFrame

    Returns:
        List of IDs of parameters selected for optimization.
    """
    return list(parameter_df.index[parameter_df[ESTIMATE] == 1])


def get_optimization_parameter_scaling(
    parameter_df: pd.DataFrame,
) -> Dict[str, str]:
    """
    Get Dictionary with optimization parameter IDs mapped to parameter scaling
    strings.

    Arguments:
        parameter_df: PEtab parameter DataFrame

    Returns:
        Dictionary with optimization parameter IDs mapped to parameter scaling
        strings.
    """
    estimated_df = parameter_df.loc[parameter_df[ESTIMATE] == 1]
    return dict(zip(estimated_df.index, estimated_df[PARAMETER_SCALE]))


def create_parameter_df(
    sbml_model: Optional[libsbml.Model] = None,
    condition_df: Optional[pd.DataFrame] = None,
    observable_df: Optional[pd.DataFrame] = None,
    measurement_df: Optional[pd.DataFrame] = None,
    model: Optional[Model] = None,
    include_optional: bool = False,
    parameter_scale: str = LOG10,
    lower_bound: Iterable = None,
    upper_bound: Iterable = None,
    mapping_df: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """Create a new PEtab parameter table

    All table entries can be provided as string or list-like with length
    matching the number of parameters

    Arguments:
        sbml_model: SBML Model (deprecated, mutually exclusive with ``model``)
        model: PEtab model (mutually exclusive with ``sbml_model``)
        condition_df: PEtab condition DataFrame
        observable_df: PEtab observable DataFrame
        measurement_df: PEtab measurement DataFrame
        include_optional: By default this only returns parameters that are
            required to be present in the parameter table. If set to ``True``,
            this returns all parameters that are allowed to be present in the
            parameter table (i.e. also including parameters specified in the
            model).
        parameter_scale: parameter scaling
        lower_bound: lower bound for parameter value
        upper_bound: upper bound for parameter value
        mapping_df: PEtab mapping DataFrame

    Returns:
        The created parameter DataFrame
    """
    if sbml_model:
        warnings.warn(
            "Passing a model via the `sbml_model` argument is "
            "deprecated, use `model=petab.models.sbml_model."
            "SbmlModel(...)` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from petab.models.sbml_model import SbmlModel

        if model:
            raise ValueError(
                "Arguments `model` and `sbml_model` are " "mutually exclusive."
            )
        model = SbmlModel(sbml_model=sbml_model)
    if include_optional:
        parameter_ids = list(
            get_valid_parameters_for_parameter_table(
                model=model,
                condition_df=condition_df,
                observable_df=observable_df,
                measurement_df=measurement_df,
            )
        )
    else:
        parameter_ids = list(
            get_required_parameters_for_parameter_table(
                model=model,
                condition_df=condition_df,
                observable_df=observable_df,
                measurement_df=measurement_df,
                mapping_df=mapping_df,
            )
        )

    df = pd.DataFrame(
        data={
            PARAMETER_ID: parameter_ids,
            PARAMETER_NAME: parameter_ids,
            PARAMETER_SCALE: parameter_scale,
            LOWER_BOUND: lower_bound,
            UPPER_BOUND: upper_bound,
            NOMINAL_VALUE: np.nan,
            ESTIMATE: 1,
            INITIALIZATION_PRIOR_TYPE: "",
            INITIALIZATION_PRIOR_PARAMETERS: "",
            OBJECTIVE_PRIOR_TYPE: "",
            OBJECTIVE_PRIOR_PARAMETERS: "",
        }
    )
    df.set_index([PARAMETER_ID], inplace=True)

    # For model parameters, set nominal values as defined in the model
    for parameter_id in df.index:
        try:
            df.loc[parameter_id, NOMINAL_VALUE] = model.get_parameter_value(
                parameter_id
            )
        except ValueError:
            # parameter was introduced as condition-specific override and
            # is potentially not present in the model
            pass
    return df


def get_required_parameters_for_parameter_table(
    model: Model,
    condition_df: pd.DataFrame,
    observable_df: pd.DataFrame,
    measurement_df: pd.DataFrame,
    mapping_df: pd.DataFrame = None,
) -> Set[str]:
    """
    Get set of parameters which need to go into the parameter table

    Arguments:
        model: PEtab model
        condition_df: PEtab condition table
        observable_df: PEtab observable table
        measurement_df: PEtab measurement table
        mapping_df: PEtab mapping table

    Returns:
        Set of parameter IDs which PEtab requires to be present in the
        parameter table. That is all {observable,noise}Parameters from the
        measurement table as well as all parametric condition table overrides
        that are not defined in the model.
    """
    # use ordered dict as proxy for ordered set
    parameter_ids = OrderedDict()

    # Add parameters from measurement table, unless they are fixed parameters
    def append_overrides(overrides):
        for p in overrides:
            if isinstance(p, str) and p not in condition_df.columns:
                parameter_ids[p] = None

    for _, row in measurement_df.iterrows():
        # we trust that the number of overrides matches
        append_overrides(
            measurements.split_parameter_replacement_list(
                row.get(OBSERVABLE_PARAMETERS, None)
            )
        )
        append_overrides(
            measurements.split_parameter_replacement_list(
                row.get(NOISE_PARAMETERS, None)
            )
        )

    # Add output parameters except for placeholders
    for formula_type, placeholder_sources in (
        (
            # Observable formulae
            {'observables': True, 'noise': False},
            # can only contain observable placeholders
            {'noise': False, 'observables': True}
        ),
        (
            # Noise formulae
            {'observables': False, 'noise': True},
            # can contain noise and observable placeholders
            {'noise': True, 'observables': True}
        ),
    ):
        output_parameters = observables.get_output_parameters(
            observable_df, model, mapping_df=mapping_df, **formula_type,
        )
        placeholders = observables.get_placeholders(
            observable_df,
            **placeholder_sources,
        )
        for p in output_parameters:
            if p not in placeholders:
                parameter_ids[p] = None

    # Add condition table parametric overrides unless already defined in the
    #  model
    for p in conditions.get_parametric_overrides(condition_df):
        if not model.has_entity_with_id(p):
            parameter_ids[p] = None

    # remove parameters that occur in the condition table and are overridden
    #  for ALL conditions
    for p in condition_df.columns[~condition_df.isnull().any()]:
        try:
            del parameter_ids[p]
        except KeyError:
            pass
    return parameter_ids.keys()


def get_valid_parameters_for_parameter_table(
    model: Model,
    condition_df: pd.DataFrame,
    observable_df: pd.DataFrame,
    measurement_df: pd.DataFrame,
    mapping_df: pd.DataFrame = None,
) -> Set[str]:
    """
    Get set of parameters which may be present inside the parameter table

    Arguments:
        model: PEtab model
        condition_df: PEtab condition table
        observable_df: PEtab observable table
        measurement_df: PEtab measurement table
        mapping_df: PEtab mapping table for additional checks

    Returns:
        Set of parameter IDs which PEtab allows to be present in the
        parameter table.
    """
    # - grab all allowed model parameters
    # - grab corresponding names from mapping table
    # - grab all output parameters defined in {observable,noise}Formula
    # - grab all parameters from measurement table
    # - grab all parametric overrides from condition table
    # - remove parameters for which condition table columns exist
    # - remove placeholder parameters
    #   (only partial overrides are not supported)

    # must not go into parameter table
    blackset = set()

    if observable_df is not None:
        placeholders = set(observables.get_placeholders(observable_df))

        # collect assignment targets
        blackset |= placeholders

    if condition_df is not None:
        blackset |= set(condition_df.columns.values) - {CONDITION_NAME}

    # don't use sets here, to have deterministic ordering,
    #  e.g. for creating parameter tables
    parameter_ids = OrderedDict.fromkeys(
        p
        for p in model.get_valid_parameters_for_parameter_table()
        if p not in blackset
    )

    if mapping_df is not None:
        for from_id, to_id in zip(
            mapping_df.index.values, mapping_df[MODEL_ENTITY_ID]
        ):
            if to_id in parameter_ids.keys():
                parameter_ids[from_id] = None

    if observable_df is not None:
        # add output parameters from observables table
        output_parameters = observables.get_output_parameters(
            observable_df=observable_df, model=model
        )
        for p in output_parameters:
            if p not in blackset:
                parameter_ids[p] = None

    # Append parameters from measurement table, unless they occur as condition
    # table columns
    def append_overrides(overrides):
        for p in overrides:
            if isinstance(p, str) and p not in blackset:
                parameter_ids[p] = None

    if measurement_df is not None:
        for _, row in measurement_df.iterrows():
            # we trust that the number of overrides matches
            append_overrides(
                measurements.split_parameter_replacement_list(
                    row.get(OBSERVABLE_PARAMETERS, None)
                )
            )
            append_overrides(
                measurements.split_parameter_replacement_list(
                    row.get(NOISE_PARAMETERS, None)
                )
            )

    # Append parameter overrides from condition table
    if condition_df is not None:
        for p in conditions.get_parametric_overrides(condition_df):
            parameter_ids[p] = None

    return parameter_ids.keys()


def get_priors_from_df(
    parameter_df: pd.DataFrame,
    mode: Literal["initialization", "objective"],
    parameter_ids: Sequence[str] = None,
) -> List[Tuple]:
    """Create list with information about the parameter priors

    Arguments:
        parameter_df: PEtab parameter table
        mode: ``'initialization'`` or ``'objective'``
        parameter_ids: A sequence of parameter IDs for which to sample starting points.
            For subsetting or reordering the parameters.
            Defaults to all estimated parameters.

    Returns:
        List with prior information.
    """
    # get types and parameters of priors from dataframe
    par_to_estimate = parameter_df.loc[parameter_df[ESTIMATE] == 1]

    if parameter_ids:
        try:
            par_to_estimate = par_to_estimate.loc[parameter_ids, :]
        except KeyError as e:
            missing_ids = set(parameter_ids) - set(par_to_estimate.index)
            raise KeyError(
                f"Parameter table does not contain estimated parameter(s) {missing_ids}."
            ) from e

    prior_list = []
    for _, row in par_to_estimate.iterrows():
        # retrieve info about type
        prior_type = str(row.get(f"{mode}PriorType", ""))
        if core.is_empty(prior_type):
            prior_type = PARAMETER_SCALE_UNIFORM

        # retrieve info about parameters of priors, make it a tuple of floats
        pars_str = str(row.get(f"{mode}PriorParameters", ""))
        if core.is_empty(pars_str):
            lb, ub = map_scale(
                [row[LOWER_BOUND], row[UPPER_BOUND]],
                [row[PARAMETER_SCALE]] * 2,
            )
            pars_str = f"{lb}{PARAMETER_SEPARATOR}{ub}"
        prior_pars = tuple(
            float(entry) for entry in pars_str.split(PARAMETER_SEPARATOR)
        )

        # add parameter scale and bounds, as this may be needed
        par_scale = row[PARAMETER_SCALE]
        par_bounds = (row[LOWER_BOUND], row[UPPER_BOUND])

        # if no prior is specified, we assume a non-informative (uniform) one
        if prior_type == "nan":
            prior_type = PARAMETER_SCALE_UNIFORM
            prior_pars = (
                scale(row[LOWER_BOUND], par_scale),
                scale(row[UPPER_BOUND], par_scale),
            )

        prior_list.append((prior_type, prior_pars, par_scale, par_bounds))

    return prior_list


def scale(
    parameter: numbers.Number,
    scale_str: PARAMETER_SCALE_ARGS,
) -> numbers.Number:
    """Scale parameter according to ``scale_str``.

    Arguments:
        parameter:
            Parameter to be scaled.
        scale_str:
            One of ``'lin'`` (synonymous with ``''``), ``'log'``, ``'log10'``.

    Returns:
        The scaled parameter.
    """
    if scale_str == LIN or not scale_str:
        return parameter
    if scale_str == LOG:
        return np.log(parameter)
    if scale_str == LOG10:
        return np.log10(parameter)
    raise ValueError(f"Invalid parameter scaling: {scale_str}")


def unscale(
    parameter: numbers.Number,
    scale_str: PARAMETER_SCALE_ARGS,
) -> numbers.Number:
    """Unscale parameter according to ``scale_str``.

    Arguments:
        parameter:
            Parameter to be unscaled.
        scale_str:
            One of ``'lin'`` (synonymous with ``''``), ``'log'``, ``'log10'``.

    Returns:
        The unscaled parameter.
    """
    if scale_str == LIN or not scale_str:
        return parameter
    if scale_str == LOG:
        return np.exp(parameter)
    if scale_str == LOG10:
        return 10**parameter
    raise ValueError(f"Invalid parameter scaling: {scale_str}")


def map_scale(
    parameters: Sequence[numbers.Number],
    scale_strs: Union[Iterable[PARAMETER_SCALE_ARGS], PARAMETER_SCALE_ARGS],
) -> Iterable[numbers.Number]:
    """Scale the parameters, i.e. as :func:`scale`, but for Sequences.

    Arguments:
        parameters:
            Parameters to be scaled.
        scale_strs:
            Scales to apply. Broadcast if a single string.

    Returns:
        The scaled parameters.
    """
    if isinstance(scale_strs, str):
        scale_strs = [scale_strs] * len(parameters)
    return map(lambda x: scale(x[0], x[1]), zip(parameters, scale_strs))


def map_unscale(
    parameters: Sequence[numbers.Number],
    scale_strs: Union[Iterable[PARAMETER_SCALE_ARGS], PARAMETER_SCALE_ARGS],
) -> Iterable[numbers.Number]:
    """Unscale the parameters, i.e. as :func:`unscale`, but for Sequences.

    Arguments:
        parameters:
            Parameters to be unscaled.
        scale_strs:
            Scales that the parameters are currently on.
            Broadcast if a single string.

    Returns:
        The unscaled parameters.
    """
    if isinstance(scale_strs, str):
        scale_strs = [scale_strs] * len(parameters)
    return map(lambda x: unscale(x[0], x[1]), zip(parameters, scale_strs))


def normalize_parameter_df(parameter_df: pd.DataFrame) -> pd.DataFrame:
    """Add missing columns and fill in default values."""
    df = parameter_df.copy(deep=True)

    if PARAMETER_NAME not in df:
        df[PARAMETER_NAME] = df.reset_index()[PARAMETER_ID]

    prior_type_cols = [INITIALIZATION_PRIOR_TYPE, OBJECTIVE_PRIOR_TYPE]
    prior_par_cols = [
        INITIALIZATION_PRIOR_PARAMETERS,
        OBJECTIVE_PRIOR_PARAMETERS,
    ]
    # iterate over initialization and objective priors
    for prior_type_col, prior_par_col in zip(prior_type_cols, prior_par_cols):
        # fill in default values for prior type
        if prior_type_col not in df:
            df[prior_type_col] = PARAMETER_SCALE_UNIFORM
        else:
            for irow, row in df.iterrows():
                if core.is_empty(row[prior_type_col]):
                    df.loc[irow, prior_type_col] = PARAMETER_SCALE_UNIFORM
        if prior_par_col not in df:
            df[prior_par_col] = None
        for irow, row in df.iterrows():
            if (
                core.is_empty(row[prior_par_col])
                and row[prior_type_col] == PARAMETER_SCALE_UNIFORM
            ):
                lb, ub = map_scale(
                    [row[LOWER_BOUND], row[UPPER_BOUND]],
                    [row[PARAMETER_SCALE]] * 2,
                )
                df.loc[irow, prior_par_col] = f"{lb}{PARAMETER_SEPARATOR}{ub}"

    return df



"""Tests for petab/simulate.py."""
import functools
from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
import pytest
import scipy.stats

import petab
from petab.C import MEASUREMENT


class TestSimulator(petab.simulate.Simulator):
    """Dummy simulator."""

    __test__ = False

    def simulate_without_noise(self) -> pd.DataFrame:
        """Dummy simulation."""
        return self.petab_problem.measurement_df


@pytest.fixture
def petab_problem() -> petab.Problem:
    """Create a PEtab problem for use in tests."""
    petab_yaml_path = (
        Path(__file__).parent.parent.absolute()
        / "doc"
        / "example"
        / "example_Fujita"
        / "Fujita.yaml"
    )
    return petab.Problem.from_yaml(str(petab_yaml_path))


def test_remove_working_dir(petab_problem):
    """Test creation and removal of a non-empty temporary working directory."""
    simulator = TestSimulator(petab_problem)
    # The working directory exists
    assert Path(simulator.working_dir).is_dir()
    synthetic_data_df = simulator.simulate()
    synthetic_data_df.to_csv(Path(simulator.working_dir, "test.csv"), sep="\t")
    simulator.remove_working_dir()
    # The (non-empty) working directory is removed
    assert not Path(simulator.working_dir).is_dir()

    # Test creation and removal of a specified working directory
    working_dir = Path("tests/test_simulate_working_dir")
    simulator = TestSimulator(petab_problem, working_dir=working_dir)
    # The working directory is as specified
    assert working_dir == Path(simulator.working_dir)
    # The working directory exists
    assert Path(simulator.working_dir).is_dir()
    # A user-specified working directory should not be removed unless
    # `force=True`.
    simulator.remove_working_dir()
    # The user-specified working directory is not removed without `force=True`
    assert Path(simulator.working_dir).is_dir()
    simulator.remove_working_dir(force=True)
    # The user-specified working directory is removed with `force=True`
    assert not Path(simulator.working_dir).is_dir()

    # Test creation and removal of a specified non-empty working directory
    simulator = TestSimulator(petab_problem, working_dir=working_dir)
    synthetic_data_df = simulator.simulate()
    synthetic_data_df.to_csv(f"{simulator.working_dir}/test.csv", sep="\t")
    simulator.remove_working_dir(force=True)
    # The non-empty, user-specified directory is removed with `force=True`
    assert not Path(simulator.working_dir).is_dir()


def test_zero_bounded(petab_problem):
    """Test `zero_bounded` argument of `sample_noise`."""
    positive = np.spacing(1)
    negative = -positive

    simulator = TestSimulator(petab_problem)
    # Set the random seed to ensure predictable tests.
    simulator.rng = np.random.default_rng(seed=0)

    # Set approximately half of the measurements to negative values, and the
    # rest to positive values.
    n_measurements = len(petab_problem.measurement_df)
    neg_indices = range(round(n_measurements / 2))
    pos_indices = range(len(neg_indices), n_measurements)
    measurements = [
        negative
        if index in neg_indices
        else (positive if index in pos_indices else np.nan)
        for index in range(n_measurements)
    ]
    synthetic_data_df = simulator.simulate(as_measurement=True).assign(
        **{petab.C.MEASUREMENT: measurements}
    )
    # All measurements are non-zero
    assert (synthetic_data_df[MEASUREMENT] != 0).all()
    # No measurements are NaN
    assert not (np.isnan(synthetic_data_df[MEASUREMENT])).any()

    synthetic_data_df_with_noise = simulator.add_noise(
        synthetic_data_df,
    )
    # Both negative and positive values are returned by default.
    assert all(
        [
            (synthetic_data_df_with_noise[MEASUREMENT] <= 0).any(),
            (synthetic_data_df_with_noise[MEASUREMENT] >= 0).any(),
        ]
    )

    synthetic_data_df_with_noise = simulator.add_noise(
        synthetic_data_df,
        zero_bounded=True,
    )
    # Values with noise that are different in sign to values without noise are
    # zeroed.
    assert all(
        [
            (
                synthetic_data_df_with_noise[MEASUREMENT][neg_indices] <= 0
            ).all(),
            (
                synthetic_data_df_with_noise[MEASUREMENT][pos_indices] >= 0
            ).all(),
            (
                synthetic_data_df_with_noise[MEASUREMENT][neg_indices] == 0
            ).any(),
            (
                synthetic_data_df_with_noise[MEASUREMENT][pos_indices] == 0
            ).any(),
            (synthetic_data_df_with_noise[MEASUREMENT][neg_indices] < 0).any(),
            (synthetic_data_df_with_noise[MEASUREMENT][pos_indices] > 0).any(),
        ]
    )


def test_add_noise(petab_problem):
    """Test the noise generating method."""

    tested_noise_distributions = {"normal", "laplace"}
    assert set(petab.C.NOISE_MODELS) == tested_noise_distributions, (
        "The noise generation methods have only been tested for "
        f"{tested_noise_distributions}. Please edit this test to include this "
        "distribution in its tested distributions. The appropriate SciPy "
        "distribution will need to be added to "
        "`petab_numpy2scipy_distribution` in `_test_add_noise`."
    )

    for distribution in tested_noise_distributions:
        petab_problem.observable_df[petab.C.NOISE_DISTRIBUTION] = distribution
        _test_add_noise(petab_problem)


def _test_add_noise(petab_problem) -> None:
    """Test the noise generating method."""
    n_samples = 100
    noise_scaling_factor = 0.5
    ks_1samp_pvalue_threshold = 0.05
    minimum_fraction_above_threshold = 0.9
    petab_numpy2scipy_distribution = {
        "normal": "norm",
        "laplace": "laplace",
    }

    simulator = TestSimulator(petab_problem)
    # Set the random seed to ensure predictable tests.
    simulator.rng = np.random.default_rng(seed=0)
    synthetic_data_df = simulator.simulate(as_measurement=True)

    # Generate samples of noisy data
    samples = []
    for _ in range(n_samples):
        samples.append(
            simulator.add_noise(
                synthetic_data_df,
                noise_scaling_factor=noise_scaling_factor,
            )[MEASUREMENT]
        )
    samples = np.array(samples)

    expected_noise_values = [
        noise_scaling_factor
        * petab.calculate.evaluate_noise_formula(
            row,
            simulator.noise_formulas,
            petab_problem.parameter_df,
            row[MEASUREMENT],
        )
        for _, row in synthetic_data_df.iterrows()
    ]
    expected_noise_distributions = [
        petab_problem.observable_df.loc[row[petab.C.OBSERVABLE_ID]].get(
            petab.C.NOISE_DISTRIBUTION, petab.C.NORMAL
        )
        for _, row in synthetic_data_df.iterrows()
    ]

    def row2cdf(row, index) -> Callable:
        """
        Get and customize the appropriate cumulative distribution function from
        `scipy.stats`.

        Arguments:
            row:
                A row from a PEtab measurement table.
            index:
                The index of the row.

        Returns:
            The appropriate SciPy cumulative distribution function, setup to
            produce noisy simulated data.
        """
        return functools.partial(
            getattr(
                scipy.stats,
                petab_numpy2scipy_distribution[
                    expected_noise_distributions[index]
                ],
            ).cdf,
            loc=row[MEASUREMENT],
            scale=expected_noise_values[index],
        )

    # Test whether the distribution of the samples is equal to the expected
    # distribution, for each measurement.
    results = []
    for index, row in synthetic_data_df.iterrows():
        results.append(
            scipy.stats.ks_1samp(samples[:, index], row2cdf(row, index))
        )
    observed_fraction_above_threshold = sum(
        r.pvalue > ks_1samp_pvalue_threshold for r in results
    ) / len(results)
    # Sufficient distributions of measurement samples are sufficiently similar
    # to the expected distribution
    assert observed_fraction_above_threshold > minimum_fraction_above_threshold

    simulator.remove_working_dir()
    assert not Path(simulator.working_dir).is_dir()



"""
Functions for creating an overview report of a PEtab problem
"""

from pathlib import Path
from shutil import copyfile
from typing import Union

import pandas as pd

import petab
from petab.C import *

__all__ = ["create_report"]


def create_report(
    problem: petab.Problem, model_name: str, output_path: Union[str, Path] = ""
) -> None:
    """Create an HTML overview data / model overview report

    Arguments:
        problem: PEtab problem
        model_name: Name of the model, used for file name for report
        output_path: Output directory
    """

    template_dir = Path(__file__).absolute().parent / "templates"
    output_path = Path(output_path)
    template_file = "report.html"

    data_per_observable = get_data_per_observable(problem.measurement_df)
    num_conditions = len(problem.condition_df.index)

    # Setup template engine
    import jinja2

    template_loader = jinja2.FileSystemLoader(searchpath=template_dir)
    template_env = jinja2.Environment(loader=template_loader)
    template = template_env.get_template(template_file)

    # Render and save
    output_text = template.render(
        problem=problem,
        model_name=model_name,
        data_per_observable=data_per_observable,
        num_conditions=num_conditions,
    )
    with open(output_path / f"{model_name}.html", "w") as html_file:
        html_file.write(output_text)
    copyfile(template_dir / "mystyle.css", output_path / "mystyle.css")


def get_data_per_observable(measurement_df: pd.DataFrame) -> pd.DataFrame:
    """Get table with number of data points per observable and condition

    Arguments:
        measurement_df: PEtab measurement data frame
    Returns:
        Pivot table with number of data points per observable and condition
    """

    my_measurements = measurement_df.copy()

    index = [SIMULATION_CONDITION_ID]
    if PREEQUILIBRATION_CONDITION_ID in my_measurements:
        my_measurements[PREEQUILIBRATION_CONDITION_ID] = (
            my_measurements[PREEQUILIBRATION_CONDITION_ID]
            .astype("object")
            .fillna("", inplace=True)
        )
        index.append(PREEQUILIBRATION_CONDITION_ID)

    data_per_observable = pd.pivot_table(
        my_measurements,
        values=MEASUREMENT,
        aggfunc="count",
        index=index,
        columns=[OBSERVABLE_ID],
        fill_value=0,
    )

    # Add row and column sums
    data_per_observable.loc["SUM", :] = data_per_observable.sum(axis=0).values
    data_per_observable["SUM"] = data_per_observable.sum(axis=1).values

    data_per_observable = data_per_observable.astype(int)

    return data_per_observable



"""
Functions for plotting residuals.
"""
from pathlib import Path
from typing import Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from ..C import *
from ..calculate import calculate_residuals
from ..core import get_simulation_df
from ..problem import Problem

__all__ = ["plot_goodness_of_fit", "plot_residuals_vs_simulation"]


def plot_residuals_vs_simulation(
    petab_problem: Problem,
    simulations_df: Union[str, Path, pd.DataFrame],
    size: Optional[Tuple] = (10, 7),
    axes: Optional[Tuple[plt.Axes, plt.Axes]] = None,
) -> matplotlib.axes.Axes:
    """
    Plot residuals versus simulation values for measurements with normal noise
    assumption.

    Parameters
    ----------
    petab_problem:
        A PEtab problem.
    simulations_df:
        A simulation DataFrame in the PEtab format or path to the simulation
        output data file.
    size:
        Figure size.
    axes:
        Axis object.

    Returns
    -------
        ax: Axis object of the created plot.
    """
    if isinstance(simulations_df, (str, Path)):
        simulations_df = get_simulation_df(simulations_df)

    if NOISE_DISTRIBUTION in petab_problem.observable_df:
        if OBSERVABLE_TRANSFORMATION in petab_problem.observable_df:
            observable_ids = petab_problem.observable_df[
                (petab_problem.observable_df[NOISE_DISTRIBUTION] == NORMAL)
                & (
                    petab_problem.observable_df[OBSERVABLE_TRANSFORMATION]
                    == LIN
                )
            ].index

        else:
            observable_ids = petab_problem.observable_df[
                petab_problem.observable_df[NOISE_DISTRIBUTION] == NORMAL
            ].index
    else:
        observable_ids = petab_problem.observable_df.index

    if observable_ids.empty:
        raise ValueError(
            "Residuals plot is only applicable for normal "
            "additive noise assumption"
        )

    if axes is None:
        fig, axes = plt.subplots(
            1, 2, sharey=True, figsize=size, width_ratios=[2, 1]
        )
        fig.set_layout_engine("tight")
        fig.suptitle("Residuals")

    residual_df = calculate_residuals(
        measurement_dfs=petab_problem.measurement_df,
        simulation_dfs=simulations_df,
        observable_dfs=petab_problem.observable_df,
        parameter_dfs=petab_problem.parameter_df,
    )[0]

    normal_residuals = residual_df[
        residual_df[OBSERVABLE_ID].isin(observable_ids)
    ]
    simulations_normal = simulations_df[
        simulations_df[OBSERVABLE_ID].isin(observable_ids)
    ]

    # compare to standard normal distribution
    ks_result = stats.kstest(normal_residuals[RESIDUAL], stats.norm.cdf)

    # plot the residuals plot
    axes[0].hlines(
        y=0,
        xmin=min(simulations_normal[SIMULATION]),
        xmax=max(simulations_normal[SIMULATION]),
        ls="--",
        color="gray",
    )
    axes[0].scatter(simulations_normal[SIMULATION], normal_residuals[RESIDUAL])
    axes[0].text(
        0.15,
        0.85,
        f"Kolmogorov-Smirnov test results:\n"
        f"statistic: {ks_result[0]:.2f}\n"
        f"pvalue: {ks_result[1]:.2e} ",
        transform=axes[0].transAxes,
    )
    axes[0].set_xlabel("simulated values")
    axes[0].set_ylabel("residuals")

    # plot histogram
    axes[1].hist(
        normal_residuals[RESIDUAL], density=True, orientation="horizontal"
    )
    axes[1].set_xlabel("distribution")

    ymin, ymax = axes[0].get_ylim()
    ylim = max(abs(ymin), abs(ymax))
    axes[0].set_ylim(-ylim, ylim)
    axes[1].tick_params(
        left=False, labelleft=False, right=True, labelright=True
    )

    return axes


def plot_goodness_of_fit(
    petab_problem: Problem,
    simulations_df: Union[str, Path, pd.DataFrame],
    size: Tuple = (10, 7),
    ax: Optional[plt.Axes] = None,
) -> matplotlib.axes.Axes:
    """
    Plot goodness of fit.

    Parameters
    ----------
    petab_problem:
        A PEtab problem.
    simulations_df:
        A simulation DataFrame in the PEtab format or path to the simulation
        output data file.
    size:
        Figure size.
    ax:
        Axis object.

    Returns
    -------
        ax: Axis object of the created plot.
    """

    if isinstance(simulations_df, (str, Path)):
        simulations_df = get_simulation_df(simulations_df)

    if simulations_df is None or petab_problem.measurement_df is None:
        raise NotImplementedError(
            "Both measurements and simulation data "
            "are needed for goodness_of_fit"
        )

    residual_df = calculate_residuals(
        measurement_dfs=petab_problem.measurement_df,
        simulation_dfs=simulations_df,
        observable_dfs=petab_problem.observable_df,
        parameter_dfs=petab_problem.parameter_df,
    )[0]
    slope, intercept, r_value, p_value, std_err = stats.linregress(
        petab_problem.measurement_df["measurement"],
        simulations_df["simulation"],
    )  # x, y

    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        fig.set_layout_engine("tight")

    ax.scatter(
        petab_problem.measurement_df["measurement"],
        simulations_df["simulation"],
    )

    ax.axis("square")
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    lim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
    ax.set_xlim(lim)
    ax.set_ylim(lim)
    x = np.linspace(lim, 100)
    ax.plot(x, x, linestyle="--", color="gray")
    ax.plot(x, intercept + slope * x, "r", label="fitted line")

    mse = np.mean(np.abs(residual_df["residual"]))
    ax.text(
        0.1,
        0.70,
        f"$R^2$: {r_value**2:.2f}\n"
        f"slope: {slope:.2f}\n"
        f"intercept: {intercept:.2f}\n"
        f"pvalue: {std_err:.2e}\n"
        f"mean squared error: {mse:.2e}\n",
        transform=ax.transAxes,
    )

    ax.set_title("Goodness of fit")
    ax.set_xlabel("simulated values")
    ax.set_ylabel("measurements")
    return ax



"""
Visualize
=========

PEtab comes with visualization functionality. Those need to be imported via
``import petab.visualize``.

"""
import importlib.util

mpl_spec = importlib.util.find_spec("matplotlib")

from .plotting import DataProvider, Figure

__all__ = ["DataProvider", "Figure"]

if mpl_spec is not None:
    from .plot_data_and_simulation import (
        plot_problem,
        plot_with_vis_spec,
        plot_without_vis_spec,
    )
    from .plot_residuals import (
        plot_goodness_of_fit,
        plot_residuals_vs_simulation,
    )
    from .plotter import MPLPlotter

    __all__.extend(
        [
            "plot_without_vis_spec",
            "plot_with_vis_spec",
            "plot_problem",
            "plot_goodness_of_fit",
            "plot_residuals_vs_simulation",
            "MPLPlotter",
        ]
    )



"""Functions for handling PySB models"""

import itertools
import re
import sys
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import pysb

from . import MODEL_TYPE_PYSB
from .model import Model


def _pysb_model_from_path(pysb_model_file: Union[str, Path]) -> pysb.Model:
    """Load a pysb model module and return the :class:`pysb.Model` instance

    :param pysb_model_file: Full or relative path to the PySB model module
    :return: The pysb Model instance
    """
    pysb_model_file = Path(pysb_model_file)
    pysb_model_module_name = pysb_model_file.with_suffix("").name

    import importlib.util

    spec = importlib.util.spec_from_file_location(
        pysb_model_module_name, pysb_model_file
    )
    module = importlib.util.module_from_spec(spec)
    sys.modules[pysb_model_module_name] = module
    spec.loader.exec_module(module)

    # find a pysb.Model instance in the module
    # 1) check if module.model exists and is a pysb.Model
    model = getattr(module, "model", None)
    if model:
        return model

    # 2) check if there is any other pysb.Model instance
    for x in dir(module):
        attr = getattr(module, x)
        if isinstance(attr, pysb.Model):
            return attr

    raise ValueError(f"Could not find any pysb.Model in {pysb_model_file}.")


class PySBModel(Model):
    """PEtab wrapper for PySB models"""

    type_id = MODEL_TYPE_PYSB

    def __init__(self, model: pysb.Model, model_id: str):
        super().__init__()

        self.model = model
        self._model_id = model_id

    @staticmethod
    def from_file(filepath_or_buffer, model_id: str):
        return PySBModel(
            model=_pysb_model_from_path(filepath_or_buffer), model_id=model_id
        )

    def to_file(self, filename: [str, Path]):
        from pysb.export import export

        model_source = export(self.model, "pysb_flat")
        with open(filename, "w") as f:
            f.write(model_source)

    @property
    def model_id(self):
        return self._model_id

    @model_id.setter
    def model_id(self, model_id):
        self._model_id = model_id

    def get_parameter_ids(self) -> Iterable[str]:
        return (p.name for p in self.model.parameters)

    def get_parameter_value(self, id_: str) -> float:
        try:
            return self.model.parameters[id_].value
        except KeyError as e:
            raise ValueError(f"Parameter {id_} does not exist.") from e

    def get_free_parameter_ids_with_values(
        self,
    ) -> Iterable[Tuple[str, float]]:
        return ((p.name, p.value) for p in self.model.parameters)

    def has_entity_with_id(self, entity_id) -> bool:
        try:
            _ = self.model.components[entity_id]
            return True
        except KeyError:
            return False

    def get_valid_parameters_for_parameter_table(self) -> Iterable[str]:
        # all parameters are allowed in the parameter table
        return self.get_parameter_ids()

    def get_valid_ids_for_condition_table(self) -> Iterable[str]:
        return itertools.chain(
            self.get_parameter_ids(), self.get_compartment_ids()
        )

    def symbol_allowed_in_observable_formula(self, id_: str) -> bool:
        return id_ in (
            x.name
            for x in itertools.chain(
                self.model.parameters,
                self.model.observables,
                self.model.expressions,
            )
        )

    def is_valid(self) -> bool:
        # PySB models are always valid
        return True

    def is_state_variable(self, id_: str) -> bool:
        # If there is a component with that name, it's not a state variable
        # (there are no dynamically-sized compartments)
        if self.model.components.get(id_, None):
            return False

        # Try parsing the ID
        try:
            result = parse_species_name(id_)
        except ValueError:
            return False
        else:
            # check if the ID is plausible
            for monomer, compartment, site_config in result:
                pysb_monomer: pysb.Monomer = self.model.monomers.get(monomer)
                if pysb_monomer is None:
                    return False
                if compartment:
                    pysb_compartment = self.model.compartments.get(compartment)
                    if pysb_compartment is None:
                        return False
                for site, state in site_config.items():
                    if site not in pysb_monomer.sites:
                        return False
                    if state not in pysb_monomer.site_states[site]:
                        return False
                if set(pysb_monomer.sites) - set(site_config.keys()):
                    # There are undefined sites
                    return False
            return True

    def get_compartment_ids(self) -> Iterable[str]:
        return (compartment.name for compartment in self.model.compartments)


def parse_species_name(
    name: str,
) -> List[Tuple[str, Optional[str], Dict[str, Any]]]:
    """Parse a PySB species name

    :param name: Species name to parse
    :returns: List of species, representing complex constituents, each as
        a tuple of the monomer name, the compartment name, and a dict of sites
        mapping to site states.
    :raises ValueError: In case this is not a valid ID
    """
    if "=MultiState(" in name:
        raise NotImplementedError("MultiState is not yet supported.")

    complex_constituent_pattern = re.compile(
        r"^(?P<monomer>\w+)\((?P<site_config>.*)\)"
        r"( \*\* (?P<compartment>.*))?$"
    )
    result = []
    complex_constituents = name.split(" % ")

    for complex_constituent in complex_constituents:
        match = complex_constituent_pattern.match(complex_constituent)
        if not match:
            raise ValueError(
                f"Invalid species name: '{name}' " f"('{complex_constituent}')"
            )
        monomer = match.groupdict()["monomer"]
        site_config_str = match.groupdict()["site_config"]
        compartment = match.groupdict()["compartment"]

        site_config = {}
        for site_str in site_config_str.split(", "):
            if not site_str:
                continue
            site, config = site_str.split("=")
            if config == "None":
                config = None
            elif config.startswith("'"):
                if not config.endswith("'"):
                    raise ValueError(
                        f"Invalid species name: '{name}' " f"('{config}')"
                    )
                # strip quotes
                config = config[1:-1]
            else:
                config = int(config)
            site_config[site] = config
        result.append(
            (monomer, compartment, site_config),
        )

    return result


def pattern_from_string(string: str, model: pysb.Model) -> pysb.ComplexPattern:
    """Convert a pattern string to a Pattern instance"""
    parts = parse_species_name(string)
    patterns = []
    for part in parts:
        patterns.append(
            pysb.MonomerPattern(
                monomer=model.monomers.get(part[0]),
                compartment=model.compartments.get(part[1], None),
                site_conditions=part[2],
            )
        )

    return pysb.ComplexPattern(patterns, compartment=None)



"""Functionality related to the PEtab entity mapping table"""
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from . import lint
from .C import *  # noqa: F403
from .models import Model

__all__ = [
    "get_mapping_df",
    "write_mapping_df",
    "check_mapping_df",
]


def get_mapping_df(
    mapping_file: Union[None, str, Path, pd.DataFrame]
) -> pd.DataFrame:
    """
    Read the provided mapping file into a ``pandas.Dataframe``.

    Arguments:
        mapping_file: Name of file to read from or pandas.Dataframe

    Returns:
        Mapping DataFrame
    """
    if mapping_file is None:
        return mapping_file

    if isinstance(mapping_file, (str, Path)):
        mapping_file = pd.read_csv(
            mapping_file, sep="\t", float_precision="round_trip"
        )

    if not isinstance(mapping_file.index, pd.RangeIndex):
        mapping_file.reset_index(inplace=True)

    for col in MAPPING_DF_REQUIRED_COLS:
        if col not in mapping_file.columns:
            raise KeyError(
                f"Mapping table missing mandatory field {PETAB_ENTITY_ID}."
            )

        lint.assert_no_leading_trailing_whitespace(
            mapping_file.reset_index()[col].values, col
        )

    mapping_file.set_index([PETAB_ENTITY_ID], inplace=True)

    return mapping_file


def write_mapping_df(df: pd.DataFrame, filename: Union[str, Path]) -> None:
    """Write PEtab mapping table

    Arguments:
        df: PEtab mapping table
        filename: Destination file name
    """
    df = get_mapping_df(df)
    df.to_csv(filename, sep="\t", index=True)


def check_mapping_df(
    df: pd.DataFrame,
    model: Optional[Model] = None,
) -> None:
    """Run sanity checks on PEtab mapping table

    Arguments:
        df: PEtab mapping DataFrame
        model: Model for additional checking of parameter IDs

    Raises:
        AssertionError: in case of problems
    """
    lint._check_df(df, MAPPING_DF_REQUIRED_COLS[1:], "mapping")

    if df.index.name != PETAB_ENTITY_ID:
        raise AssertionError(
            f"Mapping table has wrong index {df.index.name}. "
            f"Expected {PETAB_ENTITY_ID}."
        )

    lint.check_ids(df.index.values, kind=PETAB_ENTITY_ID)

    if model:
        for model_entity_id in df[MODEL_ENTITY_ID]:
            if not model.has_entity_with_id(model_entity_id):
                raise AssertionError(
                    "Mapping table maps to unknown "
                    f"model entity ID {model_entity_id}."
                )


def resolve_mapping(mapping_df: Optional[pd.DataFrame], element: str) -> str:
    """Resolve mapping for a given element.

    :param element:
        Element to resolve.

    :param mapping_df:
        Mapping table.

    :return:
        Resolved element.
    """
    if mapping_df is None:
        return element
    if element in mapping_df.index:
        return mapping_df.loc[element, MODEL_ENTITY_ID]
    return element



"""Tests related to petab.calculate."""

import numpy as np
import pandas as pd
import pytest

from petab import (
    calculate_chi2,
    calculate_llh,
    calculate_residuals,
    calculate_single_llh,
)
from petab.C import *


def model_simple():
    "Simple model." ""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_a", "obs_b", "obs_b"],
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
            TIME: [0, 10, 0, 10],
            MEASUREMENT: [0, 1, 20, 22],
        }
    )

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_b"],
            OBSERVABLE_FORMULA: ["A", "B"],
            NOISE_FORMULA: [2, 3],
        }
    ).set_index([OBSERVABLE_ID])

    parameter_df = pd.DataFrame(
        data={PARAMETER_ID: ["par1", "par2"], NOMINAL_VALUE: [3, 4]}
    )

    simulation_df = measurement_df.copy(deep=True).rename(
        columns={MEASUREMENT: SIMULATION}
    )
    simulation_df[SIMULATION] = [2, 2, 19, 20]

    expected_residuals = {
        (2 - 0) / 2,
        (2 - 1) / 2,
        (19 - 20) / 3,
        (20 - 22) / 3,
    }
    expected_residuals_nonorm = {2 - 0, 2 - 1, 19 - 20, 20 - 22}
    expected_llh = (
        -0.5 * (np.array(list(expected_residuals)) ** 2).sum()
        - 0.5 * np.log(2 * np.pi * np.array([2, 2, 3, 3]) ** 2).sum()
    )

    return (
        measurement_df,
        observable_df,
        parameter_df,
        simulation_df,
        expected_residuals,
        expected_residuals_nonorm,
        expected_llh,
    )


def model_replicates():
    """Model with replicates."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_a"],
            SIMULATION_CONDITION_ID: ["c0", "c0"],
            TIME: [10, 10],
            MEASUREMENT: [0, 1],
        }
    )

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a"],
            OBSERVABLE_FORMULA: ["A"],
            NOISE_FORMULA: [2],
        }
    ).set_index([OBSERVABLE_ID])

    parameter_df = pd.DataFrame(
        data={PARAMETER_ID: ["par1", "par2"], NOMINAL_VALUE: [3, 4]}
    ).set_index([PARAMETER_ID])

    simulation_df = measurement_df.copy(deep=True).rename(
        columns={MEASUREMENT: SIMULATION}
    )
    simulation_df[SIMULATION] = [2, 2]

    expected_residuals = {(2 - 0) / 2, (2 - 1) / 2}
    expected_residuals_nonorm = {2 - 0, 2 - 1}
    expected_llh = (
        -0.5 * (np.array(list(expected_residuals)) ** 2).sum()
        - 0.5 * np.log(2 * np.pi * np.array([2, 2]) ** 2).sum()
    )

    return (
        measurement_df,
        observable_df,
        parameter_df,
        simulation_df,
        expected_residuals,
        expected_residuals_nonorm,
        expected_llh,
    )


def model_scalings():
    """Model with scalings."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_a"],
            SIMULATION_CONDITION_ID: ["c0", "c0"],
            TIME: [5, 10],
            MEASUREMENT: [0.5, 1],
        }
    )

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a"],
            OBSERVABLE_FORMULA: ["A"],
            OBSERVABLE_TRANSFORMATION: [LOG],
            NOISE_FORMULA: [2],
        }
    ).set_index([OBSERVABLE_ID])

    parameter_df = pd.DataFrame(
        data={PARAMETER_ID: ["par1", "par2"], NOMINAL_VALUE: [3, 4]}
    ).set_index([PARAMETER_ID])

    simulation_df = measurement_df.copy(deep=True).rename(
        columns={MEASUREMENT: SIMULATION}
    )
    simulation_df[SIMULATION] = [2, 3]

    expected_residuals = {
        (np.log(2) - np.log(0.5)) / 2,
        (np.log(3) - np.log(1)) / 2,
    }
    expected_residuals_nonorm = {
        np.log(2) - np.log(0.5),
        np.log(3) - np.log(1),
    }
    expected_llh = (
        -0.5 * (np.array(list(expected_residuals)) ** 2).sum()
        - 0.5
        * np.log(
            2 * np.pi * np.array([2, 2]) ** 2 * np.array([0.5, 1]) ** 2
        ).sum()
    )

    return (
        measurement_df,
        observable_df,
        parameter_df,
        simulation_df,
        expected_residuals,
        expected_residuals_nonorm,
        expected_llh,
    )


def model_non_numeric_overrides():
    """Model with non-numeric overrides."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_a"],
            SIMULATION_CONDITION_ID: ["c0", "c0"],
            TIME: [5, 10],
            MEASUREMENT: [0.5, 1],
            NOISE_PARAMETERS: ["7;8", "2;par1"],
        }
    )

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a"],
            OBSERVABLE_FORMULA: ["A"],
            OBSERVABLE_TRANSFORMATION: [LOG],
            NOISE_FORMULA: [
                "2*noiseParameter1_obs_a + "
                "noiseParameter2_obs_a + par2 + obs_a"
            ],
        }
    ).set_index([OBSERVABLE_ID])

    parameter_df = pd.DataFrame(
        data={PARAMETER_ID: ["par1", "par2"], NOMINAL_VALUE: [3, 4]}
    ).set_index([PARAMETER_ID])

    simulation_df = measurement_df.copy(deep=True).rename(
        columns={MEASUREMENT: SIMULATION}
    )
    simulation_df[SIMULATION] = [2, 3]

    expected_residuals = {
        (np.log(2) - np.log(0.5)) / (2 * 7 + 8 + 4 + np.log(2)),
        (np.log(3) - np.log(1)) / (2 * 2 + 3 + 4 + np.log(3)),
    }
    expected_residuals_nonorm = {
        np.log(2) - np.log(0.5),
        np.log(3) - np.log(1),
    }
    expected_llh = (
        -0.5 * (np.array(list(expected_residuals)) ** 2).sum()
        - 0.5
        * np.log(
            2
            * np.pi
            * np.array([2 * 7 + 8 + 4 + np.log(2), 2 * 2 + 3 + 4 + np.log(3)])
            ** 2
            * np.array([0.5, 1]) ** 2
        ).sum()
    )

    return (
        measurement_df,
        observable_df,
        parameter_df,
        simulation_df,
        expected_residuals,
        expected_residuals_nonorm,
        expected_llh,
    )


def model_custom_likelihood():
    """Model with customized likelihoods."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_b"],
            SIMULATION_CONDITION_ID: ["c0", "c0"],
            TIME: [5, 10],
            MEASUREMENT: [0.5, 2],
        }
    )

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_a", "obs_b"],
            OBSERVABLE_FORMULA: ["A", "B"],
            OBSERVABLE_TRANSFORMATION: [LOG, LIN],
            NOISE_FORMULA: [2, 1.5],
            NOISE_DISTRIBUTION: [LAPLACE, LAPLACE],
        }
    ).set_index([OBSERVABLE_ID])

    parameter_df = pd.DataFrame(
        data={PARAMETER_ID: ["par1", "par2"], NOMINAL_VALUE: [3, 4]}
    ).set_index([PARAMETER_ID])

    simulation_df = measurement_df.copy(deep=True).rename(
        columns={MEASUREMENT: SIMULATION}
    )
    simulation_df[SIMULATION] = [2, 3]

    expected_residuals = {(np.log(2) - np.log(0.5)) / 2, (3 - 2) / 1.5}
    expected_residuals_nonorm = {np.log(2) - np.log(0.5), 3 - 2}
    expected_llh = (
        -np.abs(list(expected_residuals)).sum()
        - np.log(2 * np.array([2, 1.5]) * np.array([0.5, 1])).sum()
    )

    return (
        measurement_df,
        observable_df,
        parameter_df,
        simulation_df,
        expected_residuals,
        expected_residuals_nonorm,
        expected_llh,
    )


@pytest.fixture
def models():
    """Test model collection covering different features."""
    return [
        model_simple(),
        model_replicates(),
        model_scalings(),
        model_non_numeric_overrides(),
        model_custom_likelihood(),
    ]


def test_calculate_residuals(models):  # pylint: disable=W0621
    """Test calculate.calculate_residuals."""
    for i_model, model in enumerate(models):
        print(f"Model {i_model}")
        (
            measurement_df,
            observable_df,
            parameter_df,
            simulation_df,
            expected_residuals,
            _,
            _,
        ) = model
        residual_dfs = calculate_residuals(
            measurement_df, simulation_df, observable_df, parameter_df
        )
        assert sorted(residual_dfs[0][RESIDUAL]) == pytest.approx(
            sorted(expected_residuals)
        )


def test_calculate_non_normalized_residuals(models):  # pylint: disable=W0621
    """Test calculate.calculate_residuals without normalization."""
    for i_model, model in enumerate(models):
        print(f"Model {i_model}")
        (
            measurement_df,
            observable_df,
            parameter_df,
            simulation_df,
            _,
            expected_residuals_nonorm,
            _,
        ) = model
        residual_dfs = calculate_residuals(
            measurement_df,
            simulation_df,
            observable_df,
            parameter_df,
            normalize=False,
        )
        assert sorted(residual_dfs[0][RESIDUAL]) == pytest.approx(
            sorted(expected_residuals_nonorm)
        )


def test_calculate_chi2(models):  # pylint: disable=W0621
    """Test calculate.calculate_chi2."""
    for i_model, model in enumerate(models):
        print(f"Model {i_model}")
        (
            measurement_df,
            observable_df,
            parameter_df,
            simulation_df,
            expected_residuals,
            _,
            _,
        ) = model
        chi2 = calculate_chi2(
            measurement_df, simulation_df, observable_df, parameter_df
        )

        expected = sum(np.array(list(expected_residuals)) ** 2)
        assert chi2 == pytest.approx(expected)


def test_calculate_llh(models):  # pylint: disable=W0621
    """Test calculate.calculate_llh."""
    for i_model, model in enumerate(models):
        print(f"Model {i_model}")
        (
            measurement_df,
            observable_df,
            parameter_df,
            simulation_df,
            _,
            _,
            expected_llh,
        ) = model
        llh = calculate_llh(
            measurement_df, simulation_df, observable_df, parameter_df
        )
        assert llh == pytest.approx(expected_llh) or expected_llh is None


def test_calculate_single_llh():
    """Test calculate.calculate_single_llh."""
    m, s, sigma = 5.3, 4.5, 1.6
    pi, log, log10 = np.pi, np.log, np.log10

    llh = calculate_single_llh(
        measurement=m,
        simulation=s,
        noise_value=sigma,
        noise_distribution=NORMAL,
        scale=LIN,
    )
    expected_llh = -0.5 * (((s - m) / sigma) ** 2 + log(2 * pi * sigma**2))
    assert llh == pytest.approx(expected_llh)

    llh = calculate_single_llh(
        measurement=m,
        simulation=s,
        noise_value=sigma,
        noise_distribution=NORMAL,
        scale=LOG,
    )
    expected_llh = -0.5 * (
        ((log(s) - log(m)) / sigma) ** 2 + log(2 * pi * sigma**2 * m**2)
    )
    assert llh == pytest.approx(expected_llh)

    llh = calculate_single_llh(
        measurement=m,
        simulation=s,
        noise_value=sigma,
        noise_distribution=NORMAL,
        scale=LOG10,
    )
    expected_llh = -0.5 * (
        ((log10(s) - log10(m)) / sigma) ** 2
        + log(2 * pi * sigma**2 * m**2 * log(10) ** 2)
    )
    assert llh == pytest.approx(expected_llh)

    llh = calculate_single_llh(
        measurement=m,
        simulation=s,
        noise_value=sigma,
        noise_distribution=LAPLACE,
        scale=LIN,
    )
    expected_llh = -abs((s - m) / sigma) - log(2 * sigma)
    assert llh == pytest.approx(expected_llh)

    llh = calculate_single_llh(
        measurement=m,
        simulation=s,
        noise_value=sigma,
        noise_distribution=LAPLACE,
        scale=LOG,
    )
    expected_llh = -abs((log(s) - log(m)) / sigma) - log(2 * sigma * m)
    assert llh == pytest.approx(expected_llh)

    llh = calculate_single_llh(
        measurement=m,
        simulation=s,
        noise_value=sigma,
        noise_distribution=LAPLACE,
        scale=LOG10,
    )
    expected_llh = -abs((log10(s) - log10(m)) / sigma) - log(
        2 * sigma * m * log(10)
    )
    assert llh == pytest.approx(expected_llh)



"""Test related to petab.models.model_pysb"""
import pysb
import pytest

from petab.models.pysb_model import (
    PySBModel,
    parse_species_name,
    pattern_from_string,
)


@pytest.fixture(scope="function")
def uses_pysb():
    """Cleanup PySB auto-exported symbols before and after test"""
    pysb.SelfExporter.cleanup()
    yield ()
    pysb.SelfExporter.cleanup()


def test_parse_species_name():
    assert parse_species_name("cyclin(Y='U', b=None)") == [
        ("cyclin", None, {"Y": "U", "b": None})
    ]

    assert parse_species_name("cdc2(Y='P', b=1) % cyclin(Y='P', b=1)") == [
        ("cdc2", None, {"Y": "P", "b": 1}),
        ("cyclin", None, {"Y": "P", "b": 1}),
    ]

    assert parse_species_name("A()") == [("A", None, {})]

    assert parse_species_name(
        "Bax(s1=1, s2=2, t=None) % Bax(s1=3, s2=1, t=None) % "
        "Bax(s1=2, s2=3, t=None)"
    ) == [
        ("Bax", None, {"s1": 1, "s2": 2, "t": None}),
        ("Bax", None, {"s1": 3, "s2": 1, "t": None}),
        ("Bax", None, {"s1": 2, "s2": 3, "t": None}),
    ]

    assert parse_species_name("A(b=None) ** X") == [("A", "X", {"b": None})]

    assert parse_species_name("A(b=1) ** X % B(a=1) ** X") == [
        ("A", "X", {"b": 1}),
        ("B", "X", {"a": 1}),
    ]

    # TODO: MultiState


def test_pysb_model(uses_pysb):
    model = pysb.Model()
    pysb.Compartment("c1")
    pysb.Monomer("A")
    pysb.Monomer("B", ["s"], {"s": ["a", "b"]})
    petab_model = PySBModel(model=model, model_id="test_model")

    assert petab_model.is_state_variable("A()") is True
    assert petab_model.is_state_variable("B(s='a')") is True

    # a compartment
    assert petab_model.is_state_variable("c1") is False
    # a monomer
    assert petab_model.is_state_variable("A") is False
    assert petab_model.is_state_variable("B") is False

    # not concrete
    assert petab_model.is_state_variable("B()") is False

    # non-existing compartment
    assert petab_model.is_state_variable("A() ** c2") is False

    # non-existing site
    assert petab_model.is_state_variable("A(s='a')") is False


def test_pattern_parsing(uses_pysb):
    model = pysb.Model()
    c1 = pysb.Compartment("c1")
    A = pysb.Monomer("A")
    B = pysb.Monomer("B", ["s"], {"s": ["a", "b"]})

    pattern = pysb.as_complex_pattern(A() ** c1)
    assert pattern_from_string(str(pattern), model).matches(pattern)
    assert str(pattern) == str(pattern_from_string("A() ** c1", model))

    pattern = pysb.as_complex_pattern(B(s="a") ** c1)
    assert pattern_from_string(str(pattern), model).matches(pattern)
    assert str(pattern) == str(pattern_from_string("B(s='a') ** c1", model))



"""Check that deprecated functionality raises but still works."""
import tempfile
from pathlib import Path

import pytest

import petab

from .test_petab import petab_problem  # noqa: F401
from .test_sbml import check_model, create_test_data


def test_problem_with_sbml_model():
    """Test that a problem can be correctly created from sbml model."""
    # retrieve test data
    (
        ss_model,
        condition_df,
        observable_df,
        measurement_df,
        parameter_df,
    ) = create_test_data()

    with pytest.deprecated_call():
        petab_problem = petab.Problem(  # noqa: F811
            sbml_model=ss_model.model,
            condition_df=condition_df,
            measurement_df=measurement_df,
            parameter_df=parameter_df,
        )

    with pytest.warns(
        UserWarning,
        match="An SBML rule was removed to set the component "
        "species_2 to a constant value.",
    ):
        _, condition_model = petab.get_model_for_condition(
            petab_problem, "condition_1"
        )

    check_model(condition_model)


def test_to_files_with_sbml_model(petab_problem):  # noqa: F811
    """Test problem.to_files."""
    with tempfile.TemporaryDirectory() as outdir:
        # create target files
        sbml_file = Path(outdir, "model.xml")
        condition_file = Path(outdir, "conditions.tsv")
        measurement_file = Path(outdir, "measurements.tsv")
        parameter_file = Path(outdir, "parameters.tsv")
        observable_file = Path(outdir, "observables.tsv")

        # write contents to files
        with pytest.deprecated_call():
            petab_problem.to_files(
                sbml_file=sbml_file,
                condition_file=condition_file,
                measurement_file=measurement_file,
                parameter_file=parameter_file,
                visualization_file=None,
                observable_file=observable_file,
                yaml_file=None,
            )

        # exemplarily load some
        parameter_df = petab.get_parameter_df(parameter_file)
        same_nans = parameter_df.isna() == petab_problem.parameter_df.isna()
        assert (
            ((parameter_df == petab_problem.parameter_df) | same_nans)
            .all()
            .all()
        )



"""Functionality for simplifying PEtab problems"""
from math import nan

import pandas as pd

import petab

from . import Problem
from .C import *  # noqa: F403
from .lint import lint_problem

__all__ = [
    "remove_nan_measurements",
    "remove_unused_observables",
    "remove_unused_conditions",
    "simplify_problem",
    "condition_parameters_to_parameter_table",
]


def remove_nan_measurements(problem: Problem):
    """Drop any measurements that are NaN"""
    problem.measurement_df = problem.measurement_df[
        ~problem.measurement_df[MEASUREMENT].isna()
    ]
    problem.measurement_df.reset_index(inplace=True, drop=True)


def remove_unused_observables(problem: Problem):
    """Remove observables that have no measurements"""
    measured_observables = set(problem.measurement_df[OBSERVABLE_ID].unique())
    problem.observable_df = problem.observable_df[
        problem.observable_df.index.isin(measured_observables)
    ]


def remove_unused_conditions(problem: Problem):
    """Remove conditions that have no measurements"""
    measured_conditions = set(
        problem.measurement_df[SIMULATION_CONDITION_ID].unique()
    )
    if PREEQUILIBRATION_CONDITION_ID in problem.measurement_df:
        measured_conditions |= set(
            problem.measurement_df[PREEQUILIBRATION_CONDITION_ID].unique()
        )

    problem.condition_df = problem.condition_df[
        problem.condition_df.index.isin(measured_conditions)
    ]


def simplify_problem(problem: Problem):
    if lint_problem(problem):
        raise ValueError("Invalid PEtab problem supplied.")

    remove_unused_observables(problem)
    remove_unused_conditions(problem)
    condition_parameters_to_parameter_table(problem)

    if lint_problem(problem):
        raise AssertionError("Invalid PEtab problem generated.")


def condition_parameters_to_parameter_table(problem: Problem):
    """Move parameters from the condition table to the parameters table, if
    the same parameter value is used for all conditions."""
    if (
        problem.condition_df is None
        or problem.condition_df.empty
        or problem.model is None
    ):
        return

    replacements = {}
    for parameter_id in problem.condition_df:
        if parameter_id == CONDITION_NAME:
            continue

        if problem.model.is_state_variable(parameter_id):
            # initial states can't go the parameters table
            continue

        series = problem.condition_df[parameter_id]
        value = petab.to_float_if_float(series.iloc[0])

        # same value for all conditions and no parametric overrides (str)?
        if isinstance(value, float) and len(series.unique()) == 1:
            replacements[parameter_id] = series.iloc[0]

    if not replacements:
        return

    rows = [
        {
            PARAMETER_ID: parameter_id,
            PARAMETER_SCALE: LIN,
            LOWER_BOUND: nan,
            UPPER_BOUND: nan,
            NOMINAL_VALUE: value,
            ESTIMATE: 0,
        }
        for parameter_id, value in replacements.items()
    ]
    rows = pd.DataFrame(rows)
    rows.set_index(PARAMETER_ID, inplace=True)

    if problem.parameter_df is None:
        problem.parameter_df = rows
    else:
        problem.parameter_df = pd.concat([problem.parameter_df, rows])

    problem.condition_df = problem.condition_df.drop(
        columns=replacements.keys()
    )



"""Tests related to petab.mapping"""
import tempfile

import pandas as pd
import pytest

from petab.C import *  # noqa: F403
from petab.mapping import *


def test_get_mapping_df():
    """Test parameters.get_mapping_df."""
    # Missing columns
    mapping_df = pd.DataFrame(
        data={
            PETAB_ENTITY_ID: ["e1"],
        }
    )

    with pytest.raises(KeyError):
        get_mapping_df(mapping_df)

    # check index is correct
    mapping_df = pd.DataFrame(
        data={
            PETAB_ENTITY_ID: ["e1"],
            MODEL_ENTITY_ID: ["m1"],
        }
    )
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        write_mapping_df(mapping_df, file_name)

    assert get_mapping_df(file_name).index == ["e1"]



"""PEtab visualization plotter classes"""
import os
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.axes
import matplotlib.ticker as mtick
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from ..C import *
from .plotting import DataPlot, DataProvider, DataSeries, Figure, Subplot

__all__ = ["Plotter", "MPLPlotter", "SeabornPlotter"]


class Plotter(ABC):
    """
    Plotter abstract base class.

    Attributes
    ----------

    figure:
        Figure instance that serves as a markup for the figure that
        should be generated
    data_provider:
        Data provider
    """

    def __init__(self, figure: Figure, data_provider: DataProvider):
        self.figure = figure
        self.data_provider = data_provider

    @abstractmethod
    def generate_figure(
        self, subplot_dir: Optional[str] = None
    ) -> Optional[Dict[str, plt.Subplot]]:
        pass


class MPLPlotter(Plotter):
    """
    Matplotlib wrapper
    """

    def __init__(self, figure: Figure, data_provider: DataProvider):
        super().__init__(figure, data_provider)

    @staticmethod
    def _error_column_for_plot_type_data(plot_type_data: str) -> Optional[str]:
        """Translate PEtab plotTypeData value to column name of internal
        data representation

        Parameters
        ----------
            plot_type_data: PEtab plotTypeData value (the way replicates
            should be handled)
        Returns
        -------
            Name of corresponding column
        """
        if plot_type_data == MEAN_AND_SD:
            return "sd"
        if plot_type_data == MEAN_AND_SEM:
            return "sem"
        if plot_type_data == PROVIDED:
            return "noise_model"
        return None

    def generate_lineplot(
        self,
        ax: matplotlib.axes.Axes,
        dataplot: DataPlot,
        plotTypeData: str,
        splitaxes_params: dict,
    ) -> Tuple[matplotlib.axes.Axes, matplotlib.axes.Axes]:
        """
        Generate lineplot.

        It is possible to plot only data or only simulation or both.

        Parameters
        ----------
        ax:
            Axis object.
        dataplot:
            Visualization settings for the plot.
        plotTypeData:
            Specifies how replicates should be handled.
        splitaxes_params:

        """
        simu_color = None
        (
            measurements_to_plot,
            simulations_to_plot,
        ) = self.data_provider.get_data_to_plot(
            dataplot, plotTypeData == PROVIDED
        )
        noise_col = self._error_column_for_plot_type_data(plotTypeData)

        label_base = dataplot.legendEntry

        # check if t_inf is there
        # todo: if only t_inf, adjust appearance for that case
        plot_at_t_inf = (
            measurements_to_plot is not None and measurements_to_plot.inf_point
        ) or (
            simulations_to_plot is not None and simulations_to_plot.inf_point
        )

        if (
            measurements_to_plot is not None
            and not measurements_to_plot.data_to_plot.empty
        ):
            # plotting all measurement data

            p = None
            if plotTypeData == REPLICATE:
                replicates = np.stack(
                    measurements_to_plot.data_to_plot.repl.values
                )
                # sorts according to ascending order of conditions
                cond, replicates = zip(
                    *sorted(
                        zip(
                            measurements_to_plot.conditions,
                            replicates
                        )
                    )
                )
                replicates = np.stack(replicates)

                if replicates.ndim == 1:
                    replicates = np.expand_dims(replicates, axis=1)

                # plot first replicate
                p = ax.plot(
                    cond,
                    replicates[:, 0],
                    linestyle="-.",
                    marker="x",
                    markersize=10,
                    label=label_base,
                )

                # plot other replicates with the same color
                ax.plot(
                    cond,
                    replicates[:, 1:],
                    linestyle="-.",
                    marker="x",
                    markersize=10,
                    color=p[0].get_color(),
                )

            # construct errorbar-plots: noise specified above
            else:
                # sorts according to ascending order of conditions
                scond, smean, snoise = zip(
                    *sorted(
                        zip(
                            measurements_to_plot.conditions,
                            measurements_to_plot.data_to_plot["mean"],
                            measurements_to_plot.data_to_plot[noise_col],
                        )
                    )
                )

                if np.inf in scond:
                    # remove inf point
                    scond = scond[:-1]
                    smean = smean[:-1]
                    snoise = snoise[:-1]

                if len(scond) > 0 and len(smean) > 0 and len(snoise) > 0:
                    # if only t=inf there will be nothing to plot
                    p = ax.errorbar(
                        scond,
                        smean,
                        snoise,
                        linestyle="-.",
                        marker=".",
                        label=label_base,
                    )

            # simulations should have the same colors if both measurements
            # and simulations are plotted
            simu_color = p[0].get_color() if p else None

        # construct simulation plot
        if (
            simulations_to_plot is not None
            and not simulations_to_plot.data_to_plot.empty
        ):
            # markers will be displayed only for points that have measurement
            # counterpart
            if measurements_to_plot is not None:
                meas_conditions = (
                    measurements_to_plot.conditions.to_numpy()
                    if isinstance(measurements_to_plot.conditions, pd.Series)
                    else measurements_to_plot.conditions
                )
                every = [
                    condition in meas_conditions
                    for condition in simulations_to_plot.conditions
                ]
            else:
                every = None

            # sorts according to ascending order of conditions
            xs, ys = map(
                list,
                zip(
                    *sorted(
                        zip(
                            simulations_to_plot.conditions,
                            simulations_to_plot.data_to_plot["mean"],
                        )
                    )
                ),
            )

            if np.inf in xs:
                # remove inf point
                xs = xs[:-1]
                ys = ys[:-1]
                every = every[:-1] if every else None

            if len(xs) > 0 and len(ys) > 0:
                p = ax.plot(
                    xs,
                    ys,
                    linestyle="-",
                    marker="o",
                    markevery=every,
                    label=label_base + " simulation",
                    color=simu_color,
                )
                # lines at t=inf should have the same colors also in case
                # only simulations are plotted
                simu_color = p[0].get_color()

        # plot inf points
        if plot_at_t_inf:
            ax, splitaxes_params["ax_inf"] = self._line_plot_at_t_inf(
                ax,
                plotTypeData,
                measurements_to_plot,
                simulations_to_plot,
                noise_col,
                label_base,
                splitaxes_params,
                color=simu_color,
            )

        return ax, splitaxes_params["ax_inf"]

    def generate_barplot(
        self,
        ax: "matplotlib.pyplot.Axes",
        dataplot: DataPlot,
        plotTypeData: str,
    ) -> None:
        """
        Generate barplot.

        Parameters
        ----------
        ax:
            Axis object.
        dataplot:
            Visualization settings for the plot.
        plotTypeData:
            Specifies how replicates should be handled.
        """
        # TODO: plotTypeData == REPLICATE?
        noise_col = self._error_column_for_plot_type_data(plotTypeData)

        (
            measurements_to_plot,
            simulations_to_plot,
        ) = self.data_provider.get_data_to_plot(
            dataplot, plotTypeData == PROVIDED
        )

        x_name = dataplot.legendEntry

        if simulations_to_plot:
            bar_kwargs = {
                "align": "edge",
                "width": -1 / 3,
            }
        else:
            bar_kwargs = {
                "align": "center",
                "width": 2 / 3,
            }

        color = plt.rcParams["axes.prop_cycle"].by_key()["color"][0]

        if measurements_to_plot is not None:
            ax.bar(
                x_name,
                measurements_to_plot.data_to_plot["mean"],
                yerr=measurements_to_plot.data_to_plot[noise_col],
                color=color,
                **bar_kwargs,
                label="measurement",
            )

        if simulations_to_plot is not None:
            bar_kwargs["width"] = -bar_kwargs["width"]
            ax.bar(
                x_name,
                simulations_to_plot.data_to_plot["mean"],
                color="white",
                edgecolor=color,
                **bar_kwargs,
                label="simulation",
            )

    def generate_scatterplot(
        self,
        ax: "matplotlib.pyplot.Axes",
        dataplot: DataPlot,
        plotTypeData: str,
    ) -> None:
        """
        Generate scatterplot.

        Parameters
        ----------
        ax:
            Axis object.
        dataplot:
            Visualization settings for the plot.
        plotTypeData:
            Specifies how replicates should be handled.
        """
        (
            measurements_to_plot,
            simulations_to_plot,
        ) = self.data_provider.get_data_to_plot(
            dataplot, plotTypeData == PROVIDED
        )

        if simulations_to_plot is None or measurements_to_plot is None:
            raise NotImplementedError(
                "Both measurements and simulation data "
                "are needed for scatter plots"
            )
        ax.scatter(
            measurements_to_plot.data_to_plot["mean"],
            simulations_to_plot.data_to_plot["mean"],
            label=getattr(dataplot, LEGEND_ENTRY),
        )
        self._square_plot_equal_ranges(ax)

    def generate_subplot(
        self,
        fig: matplotlib.figure.Figure,
        ax: matplotlib.axes.Axes,
        subplot: Subplot,
    ) -> None:
        """
        Generate subplot based on markup provided by subplot.

        Parameters
        ----------
        fig:
            Figure object.
        ax:
            Axis object.
        subplot:
            Subplot visualization settings.
        """
        # set yScale
        if subplot.yScale == LIN:
            ax.set_yscale("linear")
        elif subplot.yScale == LOG10:
            ax.set_yscale("log")
        elif subplot.yScale == LOG:
            ax.set_yscale("log", base=np.e)

        if subplot.plotTypeSimulation == BAR_PLOT:
            for data_plot in subplot.data_plots:
                self.generate_barplot(ax, data_plot, subplot.plotTypeData)

            # get rid of duplicate legends
            handles, labels = ax.get_legend_handles_labels()
            by_label = dict(zip(labels, handles))
            ax.legend(by_label.values(), by_label.keys())

            x_names = [x.legendEntry for x in subplot.data_plots]
            ax.set_xticks(range(len(x_names)))
            ax.set_xticklabels(x_names)

            for label in ax.get_xmajorticklabels():
                label.set_rotation(30)
                label.set_horizontalalignment("right")
        elif subplot.plotTypeSimulation == SCATTER_PLOT:
            for data_plot in subplot.data_plots:
                self.generate_scatterplot(ax, data_plot, subplot.plotTypeData)
        else:
            # set xScale
            if subplot.xScale == LIN:
                ax.set_xscale("linear")
            elif subplot.xScale == LOG10:
                ax.set_xscale("log")
            elif subplot.xScale == LOG:
                ax.set_xscale("log", base=np.e)
            # equidistant
            elif subplot.xScale == "order":
                ax.set_xscale("linear")
                # check if conditions are monotone decreasing or increasing
                if np.all(np.diff(subplot.conditions) < 0):
                    # monot. decreasing -> reverse
                    xlabel = subplot.conditions[::-1]
                    conditions = range(len(subplot.conditions))[::-1]
                    ax.set_xticks(range(len(conditions)), xlabel)
                elif np.all(np.diff(subplot.conditions) > 0):
                    xlabel = subplot.conditions
                    conditions = range(len(subplot.conditions))
                    ax.set_xticks(range(len(conditions)), xlabel)
                else:
                    raise ValueError(
                        "Error: x-conditions do not coincide, "
                        "some are mon. increasing, some "
                        "monotonically decreasing"
                    )

            splitaxes_params = self._preprocess_splitaxes(fig, ax, subplot)
            for data_plot in subplot.data_plots:
                ax, splitaxes_params["ax_inf"] = self.generate_lineplot(
                    ax,
                    data_plot,
                    subplot.plotTypeData,
                    splitaxes_params=splitaxes_params,
                )
            if splitaxes_params["ax_inf"] is not None:
                self._postprocess_splitaxes(
                    ax, splitaxes_params["ax_inf"], splitaxes_params["t_inf"]
                )

        # show 'e' as basis not 2.7... in natural log scale cases
        def ticks(y, _):
            return r"$e^{{{:.0f}}}$".format(np.log(y))

        if subplot.xScale == LOG:
            ax.xaxis.set_major_formatter(mtick.FuncFormatter(ticks))
        if subplot.yScale == LOG:
            ax.yaxis.set_major_formatter(mtick.FuncFormatter(ticks))

        if subplot.plotTypeSimulation != BAR_PLOT:
            ax.legend()
        ax.set_title(subplot.plotName)
        if subplot.xlim:
            ax.set_xlim(subplot.xlim)
        if subplot.ylim:
            ax.set_ylim(subplot.ylim)
        ax.autoscale_view()

        # Beautify plots
        ax.set_xlabel(subplot.xLabel)
        ax.set_ylabel(subplot.yLabel)

    def generate_figure(
        self,
        subplot_dir: Optional[str] = None,
        format_: str = "png",
    ) -> Optional[Dict[str, plt.Subplot]]:
        """
        Generate the full figure based on the markup in the figure attribute.

        Parameters
        ----------
        subplot_dir:
            A path to the folder where single subplots should be saved.
            PlotIDs will be taken as file names.
        format_:
            File format for the generated figure.
            (See :py:func:`matplotlib.pyplot.savefig` for supported options).

        Returns
        -------
        ax:
            Axis object of the created plot.
        None:
            In case subplots are saved to file.
        """
        if subplot_dir is None:
            # compute, how many rows and columns we need for the subplots
            num_row = int(np.round(np.sqrt(self.figure.num_subplots)))
            num_col = int(np.ceil(self.figure.num_subplots / num_row))

            fig, axes = plt.subplots(
                num_row, num_col, squeeze=False, figsize=self.figure.size
            )
            fig.set_layout_engine("tight")

            for ax in axes.flat[self.figure.num_subplots :]:
                ax.remove()

            axes = dict(
                zip([plot.plotId for plot in self.figure.subplots], axes.flat)
            )

        for subplot in self.figure.subplots:
            if subplot_dir is not None:
                fig, ax = plt.subplots(figsize=self.figure.size)
                fig.set_layout_engine("tight")
            else:
                ax = axes[subplot.plotId]

            try:
                self.generate_subplot(fig, ax, subplot)
            except Exception as e:
                raise RuntimeError(
                    f"Error plotting {getattr(subplot, PLOT_ID)}."
                ) from e

            if subplot_dir is not None:
                # TODO: why this doesn't work?
                plt.tight_layout()
                plt.savefig(
                    os.path.join(subplot_dir, f"{subplot.plotId}.{format_}")
                )
                plt.close()

        if subplot_dir is None:
            # TODO: why this doesn't work?
            plt.tight_layout()
            return axes

    @staticmethod
    def _square_plot_equal_ranges(
        ax: "matplotlib.pyplot.Axes", lim: Optional[Union[List, Tuple]] = None
    ) -> "matplotlib.pyplot.Axes":
        """
        Square plot with equal range for scatter plots.

        Returns
        -------
            Updated axis object.
        """

        ax.axis("square")

        if lim is None:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            lim = [np.min([xlim[0], ylim[0]]), np.max([xlim[1], ylim[1]])]

        ax.set_xlim(lim)
        ax.set_ylim(lim)

        # Same tick mark on x and y
        ax.yaxis.set_major_locator(ax.xaxis.get_major_locator())

        return ax

    @staticmethod
    def _line_plot_at_t_inf(
        ax: matplotlib.axes.Axes,
        plotTypeData: str,
        measurements_to_plot: DataSeries,
        simulations_to_plot: DataSeries,
        noise_col: str,
        label_base: str,
        split_axes_params: dict,
        color=None,
    ) -> Tuple[matplotlib.axes.Axes, matplotlib.axes.Axes]:
        """
        Plot data at t=inf.

        Parameters
        ----------
        ax:
            Axis object for the data corresponding to the finite timepoints.
        plotTypeData:
            The way replicates should be handled.
        measurements_to_plot:
            Measurements to plot.
        simulations_to_plot:
            Simulations to plot.
        noise_col:
            The name of the error column for plot_type_data.
        label_base:
            Label base.
        split_axes_params:
            A dictionary of split axes parameters with
            - Axis object for the data corresponding to t=inf
            - Time value that represents t=inf
            - left and right limits for the axis where the data corresponding
            to the finite timepoints is plotted
        color:
            Line color.

        Returns
        -------
        Two axis objects: for the data corresponding to the finite timepoints
        and for the data corresponding to t=inf
        """
        ax_inf = split_axes_params["ax_inf"]
        t_inf = split_axes_params["t_inf"]
        ax_finite_right_limit = split_axes_params["ax_finite_right_limit"]
        ax_left_limit = split_axes_params["ax_left_limit"]

        timepoints_inf = [
            ax_finite_right_limit,
            t_inf,
            ax_finite_right_limit
            + (ax_finite_right_limit - ax_left_limit) * 0.2,
        ]

        # plot measurements
        if measurements_to_plot is not None and measurements_to_plot.inf_point:
            measurements_data_to_plot_inf = (
                measurements_to_plot.data_to_plot.loc[np.inf]
            )

            if plotTypeData == REPLICATE:
                p = None
                if plotTypeData == REPLICATE:
                    replicates = measurements_data_to_plot_inf.repl
                    if replicates.ndim == 0:
                        replicates = np.expand_dims(replicates, axis=0)

                    # plot first replicate
                    p = ax_inf.plot(
                        timepoints_inf,
                        [replicates[0]] * 3,
                        linestyle="-.",
                        marker="x",
                        markersize=10,
                        markevery=[1],
                        label=label_base + " simulation",
                        color=color,
                    )

                    # plot other replicates with the same color
                    ax_inf.plot(
                        timepoints_inf,
                        [replicates[1:]] * 3,
                        linestyle="-.",
                        marker="x",
                        markersize=10,
                        markevery=[1],
                        color=p[0].get_color(),
                    )
            else:
                p = ax_inf.plot(
                    [timepoints_inf[0], timepoints_inf[2]],
                    [
                        measurements_data_to_plot_inf["mean"],
                        measurements_data_to_plot_inf["mean"],
                    ],
                    linestyle="-.",
                    color=color,
                )
                ax_inf.errorbar(
                    t_inf,
                    measurements_data_to_plot_inf["mean"],
                    measurements_data_to_plot_inf[noise_col],
                    linestyle="-.",
                    marker=".",
                    label=label_base + " simulation",
                    color=p[0].get_color(),
                )

            if color is None:
                # in case no color was provided from finite time points
                # plot and measurements are available corresponding
                # simulation should have the same color
                color = p[0].get_color()

        # plot simulations
        if simulations_to_plot is not None and simulations_to_plot.inf_point:
            simulations_data_to_plot_inf = (
                simulations_to_plot.data_to_plot.loc[np.inf]
            )

            if plotTypeData == REPLICATE:
                replicates = simulations_data_to_plot_inf.repl
                if replicates.ndim == 0:
                    replicates = np.expand_dims(replicates, axis=0)

                # plot first replicate
                p = ax_inf.plot(
                    timepoints_inf,
                    [replicates[0]] * 3,
                    linestyle="-",
                    marker="o",
                    markevery=[1],
                    label=label_base,
                    color=color,
                )

                # plot other replicates with the same color
                ax_inf.plot(
                    timepoints_inf,
                    [replicates[1:]] * 3,
                    linestyle="-",
                    marker="o",
                    markevery=[1],
                    color=p[0].get_color(),
                )
            else:
                ax_inf.plot(
                    timepoints_inf,
                    [simulations_data_to_plot_inf["mean"]] * 3,
                    linestyle="-",
                    marker="o",
                    markevery=[1],
                    color=color,
                )

        ax.set_xlim(right=ax_finite_right_limit)
        return ax, ax_inf

    @staticmethod
    def _postprocess_splitaxes(
        ax: matplotlib.axes.Axes, ax_inf: matplotlib.axes.Axes, t_inf: float
    ) -> None:
        """
        Postprocess the splitaxes: set axes limits, turn off unnecessary
        ticks and plot dashed lines highlighting the gap in the x axis.

        Parameters
        ----------
        ax:
            Axis object for the data corresponding to the finite timepoints.
        ax_inf:
            Axis object for the data corresponding to t=inf.
        t_inf:
            Time value that represents t=inf
        """
        ax_inf.tick_params(left=False, labelleft=False)
        ax_inf.spines["left"].set_visible(False)
        ax_inf.set_xticks([t_inf])
        ax_inf.set_xticklabels([r"$t_{\infty}$"])

        bottom, top = ax.get_ylim()
        left, right = ax.get_xlim()
        ax.spines["right"].set_visible(False)
        ax_inf.set_xlim(right, right + (right - left) * 0.2)
        d = (top - bottom) * 0.02
        ax_inf.vlines(
            x=right, ymin=bottom + d, ymax=top - d, ls="--", color="gray"
        )  # right
        ax.vlines(
            x=right, ymin=bottom + d, ymax=top - d, ls="--", color="gray"
        )  # left
        ax_inf.set_ylim(bottom, top)
        ax.set_ylim(bottom, top)

    def _preprocess_splitaxes(
        self,
        fig: matplotlib.figure.Figure,
        ax: matplotlib.axes.Axes,
        subplot: Subplot,
    ) -> Dict:
        """
        Prepare splitaxes if data at t=inf should be plotted: compute left and
        right limits for the axis where the data corresponding to the finite
        timepoints will be plotted, compute time point that will represent
        t=inf on the plot, create additional axes for plotting data at t=inf.
        """

        def check_data_to_plot(
            data_to_plot: DataSeries,
        ) -> Tuple[bool, Optional[float], float]:
            """
            Check if there is data available at t=inf and compute maximum and
            minimum finite time points that need to be plotted corresponding
            to a dataplot.
            """
            contains_inf = False
            max_finite_cond, min_cond = None, np.inf
            if data_to_plot is not None and len(data_to_plot.conditions):
                contains_inf = np.inf in data_to_plot.conditions
                finite_conditions = data_to_plot.conditions[
                    data_to_plot.conditions != np.inf
                ]
                max_finite_cond = (
                    np.max(finite_conditions)
                    if finite_conditions.size
                    else None
                )
                min_cond = min(data_to_plot.conditions)
            return contains_inf, max_finite_cond, min_cond

        splitaxes = False
        ax_inf = None
        t_inf, ax_finite_right_limit, ax_left_limit = None, None, np.inf
        for dataplot in subplot.data_plots:
            (
                measurements_to_plot,
                simulations_to_plot,
            ) = self.data_provider.get_data_to_plot(
                dataplot, subplot.plotTypeData == PROVIDED
            )

            contains_inf_m, max_finite_cond_m, min_cond_m = check_data_to_plot(
                measurements_to_plot
            )
            contains_inf_s, max_finite_cond_s, min_cond_s = check_data_to_plot(
                simulations_to_plot
            )

            if max_finite_cond_m is not None:
                ax_finite_right_limit = (
                    max(ax_finite_right_limit, max_finite_cond_m)
                    if ax_finite_right_limit is not None
                    else max_finite_cond_m
                )
            if max_finite_cond_s is not None:
                ax_finite_right_limit = (
                    max(ax_finite_right_limit, max_finite_cond_s)
                    if ax_finite_right_limit is not None
                    else max_finite_cond_s
                )

            ax_left_limit = min(ax_left_limit, min(min_cond_m, min_cond_s))
            # check if t=inf is contained in any data to be plotted on the
            # subplot
            if not splitaxes:
                splitaxes = contains_inf_m or contains_inf_s

        if splitaxes:
            # if t=inf is the only time point in measurements and simulations
            # ax_finite_right_limit will be None and ax_left_limit will be
            # equal to np.inf
            if ax_finite_right_limit is None and ax_left_limit == np.inf:
                ax_finite_right_limit = 10
                ax_left_limit = 0
            t_inf = (
                ax_finite_right_limit
                + (ax_finite_right_limit - ax_left_limit) * 0.1
            )
            # create axes for t=inf
            divider = make_axes_locatable(ax)
            ax_inf = divider.new_horizontal(size="10%", pad=0.3)
            fig.add_axes(ax_inf)

        return {
            "ax_inf": ax_inf,
            "t_inf": t_inf,
            "ax_finite_right_limit": ax_finite_right_limit,
            "ax_left_limit": ax_left_limit,
        }


class SeabornPlotter(Plotter):
    """
    Seaborn wrapper.
    """

    def __init__(self, figure: Figure, data_provider: DataProvider):
        super().__init__(figure, data_provider)

    def generate_figure(
        self, subplot_dir: Optional[str] = None
    ) -> Optional[Dict[str, plt.Subplot]]:
        pass



import os
import re
import sys

from setuptools import find_namespace_packages, setup


def read(fname):
    """Read a file."""
    return open(fname).read()


def absolute_links(txt):
    """Replace relative petab github links by absolute links."""

    raw_base = (
        "(https://raw.githubusercontent.com/petab-dev/libpetab-python/master/"
    )
    embedded_base = (
        "(https://github.com/petab-dev/libpetab-python/tree/master/"
    )
    # iterate over links
    for var in re.findall(r"\[.*?\]\((?!http).*?\)", txt):
        if re.match(r".*?.(png|svg)\)", var):
            # link to raw file
            rep = var.replace("(", raw_base)
        else:
            # link to github embedded file
            rep = var.replace("(", embedded_base)
        txt = txt.replace(var, rep)
    return txt


# Python version check
if sys.version_info < (3, 9, 0):
    sys.exit("PEtab requires at least Python version 3.9")

# read version from file
__version__ = ""
version_file = os.path.join("petab", "version.py")
# sets __version__
exec(read(version_file))  # pylint: disable=W0122 # nosec

ENTRY_POINTS = {
    "console_scripts": [
        "petablint = petab.petablint:main",
        "petab_visualize = petab.visualize.cli:_petab_visualize_main",
    ]
}

# project metadata
# noinspection PyUnresolvedReferences
setup(
    name="petab",
    version=__version__,
    description="Parameter estimation tabular data",
    long_description=absolute_links(read("README.md")),
    long_description_content_type="text/markdown",
    author="The PEtab developers",
    author_email="daniel.weindl@helmholtz-muenchen.de",
    url="https://github.com/PEtab-dev/libpetab-python",
    packages=find_namespace_packages(exclude=["doc*", "test*"]),
    install_requires=[
        "numpy>=1.15.1",
        "pandas>=1.2.0",
        "python-libsbml>=5.17.0",
        "sympy",
        "colorama",
        "pyyaml",
        "jsonschema",
    ],
    include_package_data=True,
    python_requires=">=3.9.0",
    entry_points=ENTRY_POINTS,
    extras_require={
        "tests": [
            "pytest",
            "pytest-cov",
            "simplesbml",
            "scipy",
            "pysb",
        ],
        "quality": [
            "flake8>=3.8.3",
        ],
        "reports": [
            # https://github.com/spatialaudio/nbsphinx/issues/641
            "Jinja2==3.0.3",
        ],
        "combine": [
            "python-libcombine>=0.2.6",
        ],
        "doc": [
            "sphinx>=3.5.3, !=5.1.0",
            "sphinxcontrib-napoleon>=0.7",
            "sphinx-markdown-tables>=0.0.15",
            "sphinx-rtd-theme>=0.5.1",
            "m2r2",
            "myst-nb>=0.14.0",
            # https://github.com/spatialaudio/nbsphinx/issues/687#issuecomment-1339271312
            "ipython>=7.21.0, !=8.7.0",
        ],
        "vis": ["matplotlib>=3.6.0", "seaborn", "scipy"],
    },
)



import os
import subprocess
from math import nan
from unittest.mock import patch

import pandas as pd
import pytest

import petab
from petab import lint
from petab.C import *

# import fixtures
pytest_plugins = [
    "tests.test_petab",
]


def test_assert_measured_observables_present():
    # create test model

    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["non-existing1"],
        }
    )

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1"],
        }
    )
    observable_df.set_index(OBSERVABLE_ID, inplace=True)

    with pytest.raises(AssertionError):
        lint.assert_measured_observables_defined(measurement_df, observable_df)


def test_condition_table_is_parameter_free():
    with patch(
        "petab.get_parametric_overrides"
    ) as mock_get_parametric_overrides:
        mock_get_parametric_overrides.return_value = []
        assert lint.condition_table_is_parameter_free(pd.DataFrame()) is True
        mock_get_parametric_overrides.assert_called_once()

        mock_get_parametric_overrides.reset_mock()
        mock_get_parametric_overrides.return_value = ["p1"]
        assert lint.condition_table_is_parameter_free(pd.DataFrame()) is False
        mock_get_parametric_overrides.assert_called_once()


def test_measurement_table_has_timepoint_specific_mappings():
    # Ensure we fail if we have time-point specific assignments

    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs1"],
            SIMULATION_CONDITION_ID: ["condition1", "condition1"],
            PREEQUILIBRATION_CONDITION_ID: [nan, nan],
            TIME: [1.0, 2.0],
            OBSERVABLE_PARAMETERS: ["obsParOverride", ""],
            NOISE_PARAMETERS: ["1.0", 1.0],
        }
    )

    assert (
        lint.measurement_table_has_timepoint_specific_mappings(measurement_df)
        is True
    )

    # both measurements different anyways
    measurement_df.loc[1, OBSERVABLE_ID] = "obs2"
    assert (
        lint.measurement_table_has_timepoint_specific_mappings(measurement_df)
        is False
    )

    # mixed numeric string
    measurement_df.loc[1, OBSERVABLE_ID] = "obs1"
    measurement_df.loc[1, OBSERVABLE_PARAMETERS] = "obsParOverride"
    assert (
        lint.measurement_table_has_timepoint_specific_mappings(measurement_df)
        is False
    )

    # different numeric values
    measurement_df.loc[1, NOISE_PARAMETERS] = 2.0
    assert (
        lint.measurement_table_has_timepoint_specific_mappings(measurement_df)
        is True
    )
    assert (
        lint.measurement_table_has_timepoint_specific_mappings(
            measurement_df, allow_scalar_numeric_noise_parameters=True
        )
        is False
    )


def test_observable_table_has_nontrivial_noise_formula():
    # Ensure we fail if we have nontrivial noise formulas

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: [
                "0obsPar1noisePar",
                "2obsPar0noisePar",
                "3obsPar0noisePar",
            ],
            OBSERVABLE_FORMULA: ["1.0", "1.0", "1.0"],
            NOISE_FORMULA: [
                "noiseParameter1_0obsPar1noisePar + 3.0",
                1e18,
                "1e18",
            ],
        }
    )

    assert (
        lint.observable_table_has_nontrivial_noise_formula(observable_df)
        is True
    )

    observable_df.loc[0, NOISE_FORMULA] = "sigma1"

    assert (
        lint.observable_table_has_nontrivial_noise_formula(observable_df)
        is False
    )


def test_assert_overrides_match_parameter_count():
    # Ensure we recognize and fail if we have wrong number of overrides
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["0obsPar1noisePar", "2obsPar0noisePar"],
            OBSERVABLE_FORMULA: [
                "1.0",
                "observableParameter1_2obsPar0noisePar + "
                "observableParameter2_2obsPar0noisePar",
            ],
            NOISE_FORMULA: ["noiseParameter1_0obsPar1noisePar", "1.0"],
        }
    )
    observable_df.set_index(OBSERVABLE_ID, inplace=True)

    measurement_df_orig = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["0obsPar1noisePar", "2obsPar0noisePar"],
            SIMULATION_CONDITION_ID: ["condition1", "condition1"],
            PREEQUILIBRATION_CONDITION_ID: ["", ""],
            TIME: [1.0, 2.0],
            OBSERVABLE_PARAMETERS: ["", "override1;override2"],
            NOISE_PARAMETERS: ["noiseParOverride", ""],
        }
    )

    # valid
    petab.assert_overrides_match_parameter_count(
        measurement_df_orig, observable_df
    )

    # 0 noise parameters given, 1 expected
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[0, NOISE_PARAMETERS] = ""
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )

    # 2 noise parameters given, 1 expected
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[0, NOISE_PARAMETERS] = "noiseParOverride;oneTooMuch"
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )

    # 1 noise parameter given, 0 allowed
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[1, NOISE_PARAMETERS] = "oneTooMuch"
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )

    # 0 observable parameters given, 2 expected
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[1, OBSERVABLE_PARAMETERS] = ""
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )

    # 1 observable parameters given, 2 expected
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[1, OBSERVABLE_PARAMETERS] = "oneMissing"
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )

    # 3 observable parameters given, 2 expected
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[
        1, OBSERVABLE_PARAMETERS
    ] = "override1;override2;oneTooMuch"
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )

    # 1 observable parameters given, 0 expected
    measurement_df = measurement_df_orig.copy()
    measurement_df.loc[0, OBSERVABLE_PARAMETERS] = "oneTooMuch"
    with pytest.raises(AssertionError):
        petab.assert_overrides_match_parameter_count(
            measurement_df, observable_df
        )


def test_assert_no_leading_trailing_whitespace():
    test_df = pd.DataFrame(
        data={
            "testId": ["name1 ", "name2"],
            "testText ": [" name1", "name2"],
            "testNumeric": [1.0, 2.0],
            "testNone": None,
        }
    )

    with pytest.raises(AssertionError):
        lint.assert_no_leading_trailing_whitespace(
            test_df.columns.values, "test"
        )

    with pytest.raises(AssertionError):
        lint.assert_no_leading_trailing_whitespace(
            test_df["testId"].values, "testId"
        )

    with pytest.raises(AssertionError):
        lint.assert_no_leading_trailing_whitespace(
            test_df["testText "].values, "testText"
        )

    lint.assert_no_leading_trailing_whitespace(
        test_df["testNumeric"].values, "testNumeric"
    )

    lint.assert_no_leading_trailing_whitespace(
        test_df["testNone"].values, "testNone"
    )


def test_assert_model_parameters_in_condition_or_parameter_table():
    import simplesbml

    from petab.models.sbml_model import SbmlModel

    ss_model = simplesbml.SbmlModel()
    ss_model.addParameter("parameter1", 0.0)
    ss_model.addParameter("noiseParameter1_", 0.0)
    ss_model.addParameter("observableParameter1_", 0.0)
    sbml_model = SbmlModel(sbml_model=ss_model.model)

    lint.assert_model_parameters_in_condition_or_parameter_table(
        sbml_model, pd.DataFrame(columns=["parameter1"]), pd.DataFrame()
    )

    lint.assert_model_parameters_in_condition_or_parameter_table(
        sbml_model, pd.DataFrame(), pd.DataFrame(index=["parameter1"])
    )

    with pytest.raises(AssertionError):
        lint.assert_model_parameters_in_condition_or_parameter_table(
            sbml_model,
            pd.DataFrame(columns=["parameter1"]),
            pd.DataFrame(index=["parameter1"]),
        )

    lint.assert_model_parameters_in_condition_or_parameter_table(
        sbml_model, pd.DataFrame(), pd.DataFrame()
    )

    ss_model.addAssignmentRule("parameter1", "parameter2")
    lint.assert_model_parameters_in_condition_or_parameter_table(
        sbml_model, pd.DataFrame(), pd.DataFrame()
    )


def test_assert_noise_distributions_valid():
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["0obsPar1noisePar", "2obsPar0noisePar"],
            NOISE_PARAMETERS: ["", ""],
            NOISE_DISTRIBUTION: ["", ""],
        }
    )
    observable_df.set_index([OBSERVABLE_ID], inplace=True)

    lint.assert_noise_distributions_valid(observable_df)

    observable_df[OBSERVABLE_TRANSFORMATION] = [LIN, LOG]
    observable_df[NOISE_DISTRIBUTION] = [NORMAL, ""]
    lint.assert_noise_distributions_valid(observable_df)

    observable_df[NOISE_DISTRIBUTION] = ["Normal", ""]
    with pytest.raises(ValueError):
        lint.assert_noise_distributions_valid(observable_df)

    observable_df.drop(columns=NOISE_DISTRIBUTION, inplace=True)
    lint.assert_noise_distributions_valid(observable_df)


def test_check_measurement_df():
    """Check measurement (and observable) tables"""
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["0obsPar1noisePar", "2obsPar0noisePar"],
            OBSERVABLE_FORMULA: ["", ""],
            NOISE_FORMULA: ["", ""],
            NOISE_DISTRIBUTION: ["", ""],
        }
    )
    observable_df.set_index([OBSERVABLE_ID], inplace=True)

    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["0obsPar1noisePar", "2obsPar0noisePar"],
            SIMULATION_CONDITION_ID: ["condition1", "condition1"],
            PREEQUILIBRATION_CONDITION_ID: ["", ""],
            TIME: [1.0, 2.0],
            MEASUREMENT: [1.0, 2.0],
            OBSERVABLE_PARAMETERS: ["", ""],
            NOISE_PARAMETERS: ["", ""],
        }
    )

    lint.check_measurement_df(measurement_df, observable_df)

    # Incompatible measurement and transformation
    observable_df[OBSERVABLE_TRANSFORMATION] = [LOG, ""]
    measurement_df[MEASUREMENT] = [-1.0, 0.0]
    with pytest.raises(ValueError):
        lint.check_measurement_df(measurement_df, observable_df)


def test_check_parameter_bounds():
    lint.check_parameter_bounds(
        pd.DataFrame({LOWER_BOUND: [1], UPPER_BOUND: [2], ESTIMATE: [1]})
    )

    with pytest.raises(AssertionError):
        lint.check_parameter_bounds(
            pd.DataFrame({LOWER_BOUND: [3], UPPER_BOUND: [2], ESTIMATE: [1]})
        )

    with pytest.raises(AssertionError):
        lint.check_parameter_bounds(
            pd.DataFrame(
                {
                    LOWER_BOUND: [-1],
                    UPPER_BOUND: [2],
                    ESTIMATE: [1],
                    PARAMETER_SCALE: [LOG10],
                }
            )
        )

    with pytest.raises(AssertionError):
        lint.check_parameter_bounds(
            pd.DataFrame(
                {
                    LOWER_BOUND: [-1],
                    UPPER_BOUND: [2],
                    ESTIMATE: [1],
                    PARAMETER_SCALE: [LOG],
                }
            )
        )


def test_assert_parameter_prior_type_is_valid():
    """Check lint.assert_parameter_prior_type_is_valid."""
    lint.assert_parameter_prior_type_is_valid(
        pd.DataFrame(
            {
                INITIALIZATION_PRIOR_TYPE: [UNIFORM, LAPLACE, ""],
                OBJECTIVE_PRIOR_TYPE: [NORMAL, LOG_NORMAL, ""],
            }
        )
    )
    lint.assert_parameter_prior_type_is_valid(pd.DataFrame())

    with pytest.raises(AssertionError):
        lint.assert_parameter_prior_type_is_valid(
            pd.DataFrame({INITIALIZATION_PRIOR_TYPE: ["normel"]})
        )


def test_assert_parameter_prior_parameters_are_valid():
    """Check lint.assert_parameter_prior_parameters_are_valid."""
    parameter_df = pd.DataFrame(
        {
            INITIALIZATION_PRIOR_TYPE: [UNIFORM, "", ""],
            INITIALIZATION_PRIOR_PARAMETERS: ["0;1", "10;20", ""],
            OBJECTIVE_PRIOR_PARAMETERS: ["0;20", "10;20", ""],
        }
    )

    lint.assert_parameter_prior_parameters_are_valid(parameter_df)

    with pytest.raises(AssertionError):
        lint.assert_parameter_prior_parameters_are_valid(
            pd.DataFrame({INITIALIZATION_PRIOR_TYPE: [NORMAL]})
        )

    with pytest.raises(AssertionError):
        lint.assert_parameter_prior_parameters_are_valid(
            pd.DataFrame({OBJECTIVE_PRIOR_PARAMETERS: ["0;1;2"]})
        )


def test_petablint_succeeds():
    """Run petablint and ensure we exit successfully for a file that should
    contain no errors"""
    dir_isensee = "../doc/example/example_Isensee/"
    dir_fujita = "../doc/example/example_Fujita/"

    # run with measurement file
    script_path = os.path.abspath(os.path.dirname(__file__))
    measurement_file = os.path.join(
        script_path, dir_isensee, "Isensee_measurementData.tsv"
    )
    result = subprocess.run(["petablint", "-m", measurement_file])
    assert result.returncode == 0

    # run with yaml
    yaml_file = os.path.join(script_path, dir_fujita, "Fujita.yaml")
    result = subprocess.run(["petablint", "-v", "-y", yaml_file])
    assert result.returncode == 0

    parameter_file = os.path.join(
        script_path, dir_fujita, "Fujita_parameters.tsv"
    )
    result = subprocess.run(["petablint", "-v", "-p", parameter_file])
    assert result.returncode == 0


def test_assert_measurement_conditions_present_in_condition_table():
    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            CONDITION_NAME: ["", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    )
    condition_df.set_index(CONDITION_ID, inplace=True)

    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["", ""],
            SIMULATION_CONDITION_ID: ["condition1", "condition1"],
            TIME: [1.0, 2.0],
            MEASUREMENT: [1.0, 2.0],
            OBSERVABLE_PARAMETERS: ["", ""],
            NOISE_PARAMETERS: ["", ""],
        }
    )

    # check we can handle missing preeq condition
    lint.assert_measurement_conditions_present_in_condition_table(
        measurement_df=measurement_df, condition_df=condition_df
    )

    # check we can handle preeq condition
    measurement_df[PREEQUILIBRATION_CONDITION_ID] = [
        "condition1",
        "condition2",
    ]

    lint.assert_measurement_conditions_present_in_condition_table(
        measurement_df=measurement_df, condition_df=condition_df
    )

    # check we detect missing condition
    measurement_df[PREEQUILIBRATION_CONDITION_ID] = [
        "missing_condition1",
        "missing_condition2",
    ]
    with pytest.raises(AssertionError):
        lint.assert_measurement_conditions_present_in_condition_table(
            measurement_df=measurement_df, condition_df=condition_df
        )


def test_check_condition_df():
    """Check that we correctly detect errors in condition table"""
    import simplesbml

    from petab.models.sbml_model import SbmlModel

    ss_model = simplesbml.SbmlModel()
    model = SbmlModel(sbml_model=ss_model.model)
    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1"],
            "p1": [nan],
        }
    )
    condition_df.set_index(CONDITION_ID, inplace=True)

    # parameter missing in model
    with pytest.raises(AssertionError):
        lint.check_condition_df(condition_df, model)

    # fix by adding output parameter
    observable_df = pd.DataFrame(
        {
            OBSERVABLE_ID: ["obs1"],
            OBSERVABLE_FORMULA: ["p1"],
        }
    )
    lint.check_condition_df(condition_df, model, observable_df)

    # fix by adding parameter
    ss_model.addParameter("p1", 1.0)
    lint.check_condition_df(condition_df, model)

    # species missing in model
    condition_df["s1"] = [3.0]
    with pytest.raises(AssertionError):
        lint.check_condition_df(condition_df, model)

    # fix:
    ss_model.addSpecies("[s1]", 1.0)
    lint.check_condition_df(condition_df, model)

    # compartment missing in model
    condition_df["c2"] = [4.0]
    with pytest.raises(AssertionError):
        lint.check_condition_df(condition_df, model)

    # fix:
    ss_model.addCompartment(comp_id="c2", vol=1.0)
    lint.check_condition_df(condition_df, model)


def test_check_ids():
    """Test check_ids"""

    lint.check_ids(["a1", "_1"])

    with pytest.raises(ValueError):
        lint.check_ids(["1"])


def test_check_parameter_df():
    """Check lint.check_parameter_df."""

    parameter_df = pd.DataFrame(
        {
            PARAMETER_ID: ["par0", "par1", "par2"],
            PARAMETER_SCALE: [LOG10, LOG10, LIN],
            NOMINAL_VALUE: [1e-2, 1e-3, 1e-4],
            ESTIMATE: [1, 1, 0],
            LOWER_BOUND: [1e-5, 1e-6, 1e-7],
            UPPER_BOUND: [1e5, 1e6, 1e7],
        }
    ).set_index(PARAMETER_ID)
    parameter_df[NOMINAL_VALUE] = parameter_df[NOMINAL_VALUE].astype("object")
    lint.check_parameter_df(df=parameter_df)

    # NOMINAL_VALUE empty, for non-estimated parameter
    parameter_df.loc["par2", NOMINAL_VALUE] = ""
    with pytest.raises(AssertionError):
        lint.check_parameter_df(df=parameter_df)

    # NOMINAL_VALUE column missing, but non-estimated parameter
    del parameter_df[NOMINAL_VALUE]
    with pytest.raises(AssertionError):
        lint.check_parameter_df(df=parameter_df)


def test_check_observable_df():
    """Check that we correctly detect errors in observable table"""

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_FORMULA: ["x1", "x2"],
            NOISE_FORMULA: ["sigma1", "sigma2"],
        }
    ).set_index(OBSERVABLE_ID)

    lint.check_observable_df(observable_df)

    # Check that duplicated observables ids are detected
    bad_observable_df = observable_df.copy()
    bad_observable_df.index = ["obs1", "obs1"]
    with pytest.raises(AssertionError):
        lint.check_observable_df(bad_observable_df)

    # Check that missing noiseFormula is detected
    bad_observable_df = observable_df.copy()
    bad_observable_df.loc["obs1", NOISE_FORMULA] = nan
    with pytest.raises(AssertionError):
        lint.check_observable_df(bad_observable_df)


def test_condition_ids_are_unique():
    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition1"],
            "parameter1": [1.0, 2.0],
        }
    )
    condition_df.set_index(CONDITION_ID, inplace=True)

    with pytest.raises(AssertionError):
        lint.check_condition_df(condition_df)

    condition_df.index = ["condition0", "condition1"]
    condition_df.index.name = "conditionId"
    lint.check_condition_df(condition_df)


def test_parameter_ids_are_unique():
    parameter_df = pd.DataFrame(
        {
            PARAMETER_ID: ["par0", "par0"],
            PARAMETER_SCALE: [LIN, LIN],
            ESTIMATE: [1, 1],
            LOWER_BOUND: [1e-5, 1e-6],
            UPPER_BOUND: [1e5, 1e6],
        }
    ).set_index(PARAMETER_ID)

    with pytest.raises(AssertionError):
        lint.check_parameter_df(parameter_df)

    parameter_df.index = ["par0", "par1"]
    parameter_df.index.name = "parameterId"
    lint.check_parameter_df(parameter_df)



"""PEtab core functions (or functions that don't fit anywhere else)"""
import logging
import os
import re
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Union,
)
from warnings import warn

import numpy as np
import pandas as pd

from . import yaml
from .C import *  # noqa: F403

logger = logging.getLogger(__name__)
__all__ = [
    "get_simulation_df",
    "write_simulation_df",
    "get_visualization_df",
    "write_visualization_df",
    "get_notnull_columns",
    "flatten_timepoint_specific_output_overrides",
    "concat_tables",
    "to_float_if_float",
    "is_empty",
    "create_combine_archive",
    "unique_preserve_order",
    "unflatten_simulation_df",
]

POSSIBLE_GROUPVARS_FLATTENED_PROBLEM = [
    OBSERVABLE_ID,
    OBSERVABLE_PARAMETERS,
    NOISE_PARAMETERS,
    SIMULATION_CONDITION_ID,
    PREEQUILIBRATION_CONDITION_ID,
]


def get_simulation_df(simulation_file: Union[str, Path]) -> pd.DataFrame:
    """Read PEtab simulation table

    Arguments:
        simulation_file: URL or filename of PEtab simulation table

    Returns:
        Simulation DataFrame
    """
    return pd.read_csv(
        simulation_file, sep="\t", index_col=None, float_precision="round_trip"
    )


def write_simulation_df(df: pd.DataFrame, filename: Union[str, Path]) -> None:
    """Write PEtab simulation table

    Arguments:
        df: PEtab simulation table
        filename: Destination file name
    """
    df.to_csv(filename, sep="\t", index=False)


def get_visualization_df(
    visualization_file: Union[str, Path, pd.DataFrame, None]
) -> Union[pd.DataFrame, None]:
    """Read PEtab visualization table

    Arguments:
        visualization_file:
            URL or filename of PEtab visualization table to read from,
            or a DataFrame or None that will be returned as is.

    Returns:
        Visualization DataFrame
    """
    if visualization_file is None:
        return None

    if isinstance(visualization_file, pd.DataFrame):
        return visualization_file

    try:
        types = {PLOT_NAME: str}
        vis_spec = pd.read_csv(
            visualization_file,
            sep="\t",
            index_col=None,
            converters=types,
            float_precision="round_trip",
        )
    except pd.errors.EmptyDataError:
        warn(
            "Visualization table is empty. Defaults will be used. "
            "Refer to the documentation for details."
        )
        vis_spec = pd.DataFrame()
    return vis_spec


def write_visualization_df(
    df: pd.DataFrame, filename: Union[str, Path]
) -> None:
    """Write PEtab visualization table

    Arguments:
        df: PEtab visualization table
        filename: Destination file name
    """
    df.to_csv(filename, sep="\t", index=False)


def get_notnull_columns(df: pd.DataFrame, candidates: Iterable):
    """
    Return list of ``df``-columns in ``candidates`` which are not all null/nan.

    The output can e.g. be used as input for ``pandas.DataFrame.groupby``.

    Arguments:
        df:
            Dataframe
        candidates:
            Columns of ``df`` to consider
    """
    return [
        col for col in candidates if col in df and not np.all(df[col].isnull())
    ]


def get_observable_replacement_id(groupvars, groupvar) -> str:
    """Get the replacement ID for an observable.

    Arguments:
        groupvars:
            The columns of a PEtab measurement table that should be unique
            between observables in a flattened PEtab problem.
        groupvar:
            A specific grouping of `groupvars`.

    Returns:
        The observable replacement ID.
    """
    replacement_id = ""
    for field in POSSIBLE_GROUPVARS_FLATTENED_PROBLEM:
        if field in groupvars:
            val = (
                str(groupvar[groupvars.index(field)])
                .replace(PARAMETER_SEPARATOR, "_")
                .replace(".", "_")
            )
            if replacement_id == "":
                replacement_id = val
            elif val != "":
                replacement_id += f"__{val}"
    return replacement_id


def get_hyperparameter_replacement_id(
    hyperparameter_type,
    observable_replacement_id,
):
    """Get the full ID for a replaced hyperparameter.

    Arguments:
        hyperparameter_type:
            The type of hyperparameter, e.g. `noiseParameter`.
        observable_replacement_id:
            The observable replacement ID, e.g. the output of
            `get_observable_replacement_id`.

    Returns:
        The hyperparameter replacement ID, with a field that will be replaced
        by the first matched substring in a regex substitution.
    """
    return f"{hyperparameter_type}\\1_{observable_replacement_id}"


def get_flattened_id_mappings(
    petab_problem: "petab.problem.Problem",
) -> Dict[str, Dict[str, str]]:
    """Get mapping from unflattened to flattened observable IDs.

    Arguments:
        petab_problem:
            The unflattened PEtab problem.

    Returns:
        A dictionary of dictionaries. Each inner dictionary is a mapping
        from original ID to flattened ID. Each outer dictionary is the mapping
        for either: observable IDs; noise parameter IDs; or, observable
        parameter IDs.
    """
    groupvars = get_notnull_columns(
        petab_problem.measurement_df, POSSIBLE_GROUPVARS_FLATTENED_PROBLEM
    )
    mappings = {
        OBSERVABLE_ID: {},
        NOISE_PARAMETERS: {},
        OBSERVABLE_PARAMETERS: {},
    }
    for groupvar, measurements in petab_problem.measurement_df.groupby(
        groupvars, dropna=False
    ):
        observable_id = groupvar[groupvars.index(OBSERVABLE_ID)]
        observable_replacement_id = get_observable_replacement_id(
            groupvars, groupvar
        )

        logger.debug(f"Creating synthetic observable {observable_id}")
        if observable_replacement_id in petab_problem.observable_df.index:
            raise RuntimeError(
                "could not create synthetic observables "
                f"since {observable_replacement_id} was "
                "already present in observable table"
            )

        mappings[OBSERVABLE_ID][observable_replacement_id] = observable_id

        for field, hyperparameter_type, target in [
            (NOISE_PARAMETERS, "noiseParameter", NOISE_FORMULA),
            (OBSERVABLE_PARAMETERS, "observableParameter", OBSERVABLE_FORMULA),
        ]:
            if field in measurements:
                mappings[field][
                    get_hyperparameter_replacement_id(
                        hyperparameter_type=hyperparameter_type,
                        observable_replacement_id=observable_replacement_id,
                    )
                ] = rf"{hyperparameter_type}([0-9]+)_{observable_id}"
    return mappings


def flatten_timepoint_specific_output_overrides(
    petab_problem: "petab.problem.Problem",
) -> None:
    """Flatten timepoint-specific output parameter overrides.

    If the PEtab problem definition has timepoint-specific
    `observableParameters` or `noiseParameters` for the same observable,
    replace those by replicating the respective observable.

    This is a helper function for some tools which may not support such
    timepoint-specific mappings. The observable table and measurement table
    are modified in place.

    Arguments:
        petab_problem:
            PEtab problem to work on. Modified in place.
    """
    new_measurement_dfs = []
    new_observable_dfs = []
    groupvars = get_notnull_columns(
        petab_problem.measurement_df, POSSIBLE_GROUPVARS_FLATTENED_PROBLEM
    )

    mappings = get_flattened_id_mappings(petab_problem)

    for groupvar, measurements in petab_problem.measurement_df.groupby(
        groupvars, dropna=False
    ):
        obs_id = groupvar[groupvars.index(OBSERVABLE_ID)]
        observable_replacement_id = get_observable_replacement_id(
            groupvars, groupvar
        )

        observable = petab_problem.observable_df.loc[obs_id].copy()
        observable.name = observable_replacement_id
        for field, hyperparameter_type, target in [
            (NOISE_PARAMETERS, "noiseParameter", NOISE_FORMULA),
            (OBSERVABLE_PARAMETERS, "observableParameter", OBSERVABLE_FORMULA),
            (OBSERVABLE_PARAMETERS, "observableParameter", NOISE_FORMULA),
        ]:
            if field not in measurements:
                continue

            hyperparameter_replacement_id = get_hyperparameter_replacement_id(
                hyperparameter_type=hyperparameter_type,
                observable_replacement_id=observable_replacement_id,
            )
            hyperparameter_id = mappings[field][hyperparameter_replacement_id]
            observable[target] = re.sub(
                hyperparameter_id,
                hyperparameter_replacement_id,
                observable[target],
            )

        measurements[OBSERVABLE_ID] = observable_replacement_id
        new_measurement_dfs.append(measurements)
        new_observable_dfs.append(observable)

    petab_problem.observable_df = pd.concat(new_observable_dfs, axis=1).T
    petab_problem.observable_df.index.name = OBSERVABLE_ID
    petab_problem.measurement_df = pd.concat(new_measurement_dfs)


def unflatten_simulation_df(
    simulation_df: pd.DataFrame,
    petab_problem: "petab.problem.Problem",
) -> pd.DataFrame:
    """Unflatten simulations from a flattened PEtab problem.

    A flattened PEtab problem is the output of applying
    :func:`flatten_timepoint_specific_output_overrides` to a PEtab problem.

    Arguments:
        simulation_df:
            The simulation dataframe. A dataframe in the same format as a PEtab
            measurements table, but with the ``measurement`` column switched
            with a ``simulation`` column.
        petab_problem:
            The unflattened PEtab problem.

    Returns:
        The simulation dataframe for the unflattened PEtab problem.
    """
    mappings = get_flattened_id_mappings(petab_problem)
    original_observable_ids = simulation_df[OBSERVABLE_ID].replace(
        mappings[OBSERVABLE_ID]
    )
    unflattened_simulation_df = simulation_df.assign(
        **{
            OBSERVABLE_ID: original_observable_ids,
        }
    )
    return unflattened_simulation_df


def concat_tables(
    tables: Union[
        str, Path, pd.DataFrame, Iterable[Union[pd.DataFrame, str, Path]]
    ],
    file_parser: Optional[Callable] = None,
) -> pd.DataFrame:
    """Concatenate DataFrames provided as DataFrames or filenames, and a parser

    Arguments:
        tables:
            Iterable of tables to join, as DataFrame or filename.
        file_parser:
            Function used to read the table in case filenames are provided,
            accepting a filename as only argument.

    Returns:
        The concatenated DataFrames
    """

    if isinstance(tables, pd.DataFrame):
        return tables

    if isinstance(tables, (str, Path)):
        return file_parser(tables)

    df = pd.DataFrame()

    for tmp_df in tables:
        # load from file, if necessary
        if isinstance(tmp_df, (str, Path)):
            tmp_df = file_parser(tmp_df)

        df = pd.concat(
            [df, tmp_df],
            sort=False,
            ignore_index=isinstance(tmp_df.index, pd.RangeIndex),
        )

    return df


def to_float_if_float(x: Any) -> Any:
    """Return input as float if possible, otherwise return as is

    Arguments:
        x: Anything

    Returns:
        ``x`` as float if possible, otherwise ``x``
    """

    try:
        return float(x)
    except (ValueError, TypeError):
        return x


def is_empty(val) -> bool:
    """Check if the value `val`, e.g. a table entry, is empty.

    Arguments:
        val: The value to check.

    Returns:
        Whether the field is to be considered empty.
    """
    return val == "" or pd.isnull(val)


def create_combine_archive(
    yaml_file: Union[str, Path],
    filename: Union[str, Path],
    family_name: Optional[str] = None,
    given_name: Optional[str] = None,
    email: Optional[str] = None,
    organization: Optional[str] = None,
) -> None:
    """Create COMBINE archive (https://co.mbine.org/documents/archive) based
    on PEtab YAML file.

    Arguments:
        yaml_file: Path to PEtab YAML file
        filename: Destination file name
        family_name: Family name of archive creator
        given_name: Given name of archive creator
        email: E-mail address of archive creator
        organization: Organization of archive creator
    """

    path_prefix = os.path.dirname(str(yaml_file))
    yaml_config = yaml.load_yaml(yaml_file)

    # function-level import, because module-level import interfered with
    # other SWIG interfaces
    try:
        import libcombine
    except ImportError:
        raise ImportError(
            "To use PEtab's COMBINE functionality, libcombine "
            "(python-libcombine) must be installed."
        )

    def _add_file_metadata(location: str, description: str = ""):
        """Add metadata to the added file"""
        omex_description = libcombine.OmexDescription()
        omex_description.setAbout(location)
        omex_description.setDescription(description)
        omex_description.setCreated(
            libcombine.OmexDescription.getCurrentDateAndTime()
        )
        archive.addMetadata(location, omex_description)

    archive = libcombine.CombineArchive()

    # Add PEtab files and metadata
    archive.addFile(
        str(yaml_file),
        os.path.basename(yaml_file),
        "http://identifiers.org/combine.specifications/petab.version-1",
        True,
    )
    _add_file_metadata(
        location=os.path.basename(yaml_file), description="PEtab YAML file"
    )

    # Add parameter file(s) that describe a single parameter table.
    # Works for a single file name, or a list of file names.
    for parameter_subset_file in list(
        np.array(yaml_config[PARAMETER_FILE]).flat
    ):
        archive.addFile(
            os.path.join(path_prefix, parameter_subset_file),
            parameter_subset_file,
            libcombine.KnownFormats.lookupFormat("tsv"),
            False,
        )
        _add_file_metadata(
            location=parameter_subset_file, description="PEtab parameter file"
        )

    for problem in yaml_config[PROBLEMS]:
        for sbml_file in problem[SBML_FILES]:
            archive.addFile(
                os.path.join(path_prefix, sbml_file),
                sbml_file,
                libcombine.KnownFormats.lookupFormat("sbml"),
                False,
            )
            _add_file_metadata(location=sbml_file, description="SBML model")

        for field in [
            MEASUREMENT_FILES,
            OBSERVABLE_FILES,
            VISUALIZATION_FILES,
            CONDITION_FILES,
        ]:
            if field not in problem:
                continue

            for file in problem[field]:
                archive.addFile(
                    os.path.join(path_prefix, file),
                    file,
                    libcombine.KnownFormats.lookupFormat("tsv"),
                    False,
                )
                desc = field.split("_")[0]
                _add_file_metadata(
                    location=file, description=f"PEtab {desc} file"
                )

    # Add archive metadata
    description = libcombine.OmexDescription()
    description.setAbout(".")
    description.setDescription("PEtab archive")
    description.setCreated(libcombine.OmexDescription.getCurrentDateAndTime())

    # Add creator info
    creator = libcombine.VCard()
    if family_name:
        creator.setFamilyName(family_name)
    if given_name:
        creator.setGivenName(given_name)
    if email:
        creator.setEmail(email)
    if organization:
        creator.setOrganization(organization)
    description.addCreator(creator)

    archive.addMetadata(".", description)
    archive.writeToFile(str(filename))


def unique_preserve_order(seq: Sequence) -> List:
    """Return a list of unique elements in Sequence, keeping only the first
    occurrence of each element

    Parameters:
        seq: Sequence to prune

    Returns:
        List of unique elements in ``seq``
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]



"""Command-line interface for visualization."""
import argparse
from pathlib import Path

import matplotlib.pyplot as plt

from .. import Problem, get_simulation_df, get_visualization_df
from .plot_data_and_simulation import plot_problem


def _parse_cli_args():
    """Parse command-line arguments."""

    parser = argparse.ArgumentParser(
        description="Create PEtab visualizations."
    )

    parser.add_argument(
        "-y",
        "--yaml",
        dest="yaml_file_name",
        required=True,
        help="PEtab problem YAML filename",
    )
    parser.add_argument(
        "-s",
        "--simulations",
        dest="simulation_file_name",
        required=False,
        help="PEtab simulation filename",
    )
    parser.add_argument(
        "-o",
        "--output-directory",
        dest="output_directory",
        required=True,
        help="Output directory",
    )
    parser.add_argument(
        "-v",
        "--visualizations",
        required=False,
        dest="visualization_file_name",
        help="PEtab visualization specification filename",
    )
    parser.add_argument(
        "--style",
        required=False,
        dest="style_file_name",
        help="Matplotlib style file",
    )
    return parser.parse_args()


def _petab_visualize_main():
    """Entrypoint for visualization command-line interface."""
    args = _parse_cli_args()

    petab_problem = Problem.from_yaml(args.yaml_file_name)
    simulations_df = None
    if args.simulation_file_name:
        simulations_df = get_simulation_df(args.simulation_file_name)

    if args.visualization_file_name:
        petab_problem.visualization_df = get_visualization_df(
            args.visualization_file_name
        )

    if args.style_file_name:
        plt.style.use(args.style_file_name)

    # Avoid errors when plotting without X server
    plt.switch_backend("agg")

    Path(args.output_directory).mkdir(exist_ok=True, parents=True)

    plot_problem(
        petab_problem=petab_problem,
        simulations_df=simulations_df,
        subplot_dir=args.output_directory,
    )
